#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_CasualAttentionMask Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_CasualAttentionMask')
class TextEncoder_CasualAttentionMask(Pattern):
    """The TextEncoder_CasualAttentionMask pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_CasualAttentionMask' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # for text encoder self_attn/out_proj/Add
            'TextEncoder_CasualAttentionMask': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (7, 'Concat'),
                                (8, 'ConstantOfShape'), (9, 'Shape'), (10, 'ConstantOfShape'), (11, 'Trilu'),
                                (12, 'Unsqueeze'), (13, 'Cast')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (7, 'Concat')],
                               [(), (6, 'Unsqueeze'), (7, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape'), (2, 'ConstantOfShape'), (3, 'Unsqueeze')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'diffusion_casualAttentionMask/input/reshape',
                        1: 'diffusion_casualAttentionMask/input/reshape3D',
                        2: 10,
                        3: 12
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        }], [[0], 1]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            0: [0]
                        }], [[0], 1]],
                        2: [[{
                            10: [0]
                        }], [[0], 1]],
                        3: [[{
                            13: [0]
                        }], [[0], 1]],
                    },
                    'returns': [9, 12]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['TextEncoder_CasualAttentionMask'])):
            pattern_dict = pattern_mapping_config['TextEncoder_CasualAttentionMask'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_CasualAttentionMask",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('TextEncoder_CasualAttentionMask mathched...')
                logger.debug('TextEncoder_CasualAttentionMask = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    # the first new node
                    attr = OrderedDict()
                    attr['dst_shape'] = '1,-1'
                    reshape_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[reshape_node_idx].attr = attr

                    # the second new node
                    attr_2 = OrderedDict()
                    attr_2['dst_shape'] = '-1,-1,-1'
                    attr_2['dims'] = '0, 1, 1'
                    second_reshape_node_idx = model.get_node_id(new_node_names[j][1])
                    model.nodes[second_reshape_node_idx].attr = attr_2

                    # the thrid new node
                    attr_3 = OrderedDict()
                    attr_3['trilu'] = 1
                    attr_3['value'] = -10000
                    attr_3['tensor'] = 1
                    constantOfShape_node_idx = model.get_node_id(new_node_names[j][2])
                    constantOfShape_node = model.nodes[constantOfShape_node_idx]
                    constantOfShape_node.attr = attr_3

                    # the forth new node
                    second_ret_old_node = ret_old_nodes[j][1]
                    assert second_ret_old_node.op_type == 'Unsqueeze'
                    unsqueeze_node_idx = model.get_node_id(new_node_names[j][3])
                    model.nodes[unsqueeze_node_idx].attr = second_ret_old_node.attr

                    # text encoder output modify
                    last_hidden_state_idx = len(model.nodes)
                    node_name_list = []
                    for node in model.nodes:
                        idx = model.get_node_id(node.name)
                        if 'final_layer_norm' in node.name:
                            last_hidden_state_idx = idx

                        if idx > last_hidden_state_idx:
                            node_name_list.append(node.name)

                    if node_name_list != []:
                        model.remove_nodes(node_name_list)

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The AttentionBlock_QKVReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
from .. import logger


@pattern_registry(pattern_type='AttentionBlock_QKVReshape')
class AttentionBlock_QKVReshape(Pattern):
    """The AttentionBlock_QKVReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'AttentionBlock_QKVReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # for Reshape 1 - 8, Reshape 1/3/5/7, Reshape2/4/6/8
            'AttentionBlock_QKVReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Div'), (3, 'Cast'), (4, 'Cast'),
                                (5, 'Unsqueeze'), (12, 'Concat'), (13, 'Reshape')],
                               [(), (6, 'Shape'), (7, 'Gather'), (8, 'Unsqueeze'), (12, 'Concat')],
                               [(), (9, 'Shape'), (10, 'Gather'), (11, 'Unsqueeze'), (12, 'Concat')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 13,
                    },
                    'input_tensors': {
                        0: [[{
                            13: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            13: [0]
                        }], [[0], 1]]
                    },
                    'returns': [12, 13, 0]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['AttentionBlock_QKVReshape'])):
            pattern_dict = pattern_mapping_config['AttentionBlock_QKVReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("AttentionBlock_QKVReshape",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('AttentionBlock_QKVReshape mathched...')
                logger.debug('AttentionBlock_QKVReshape = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    concat_node = ret_old_nodes[j][0]
                    concat_node_input_size = len(concat_node.input_tensors)

                    if concat_node_input_size == 4:
                        # get weight from MatMul Node
                        matmul_node_name = ret_old_nodes[j][1].input_tensors[0].source_op[0]
                        matmul_node = model.get_node_by_name(matmul_node_name)
                        # sometimes matmul_node.op_type could be BatchMatmul, /decoder/mid_block/attentions.0/Reshape_7
                        if matmul_node.op_type == 'MatMulWithBias':
                            weight = matmul_node.input_tensors[1].data.shape[1]
                        elif matmul_node.op_type == 'MatMul':
                            weight = matmul_node.input_tensors[1].data.shape[1]

                        constant_weight_idx = False
                        for idx in range(len(concat_node.input_tensors)):
                            if concat_node.input_tensors[idx].data != None:
                                constant_weight_idx = idx
                        if constant_weight_idx != False:
                            dim = int(concat_node.input_tensors[constant_weight_idx].data)

                        attr = OrderedDict()
                        attr['dst_shape'] = []
                        for i in range(concat_node_input_size):
                            if i == constant_weight_idx:
                                attr['dst_shape'].append(str(dim))
                            elif i == 3:
                                attr['dst_shape'].append(str(int(weight / dim)))
                            else:
                                attr['dst_shape'].append('-1')

                        attr['dst_shape'] = ','.join(attr['dst_shape'])
                        attr['dims'] = 0
                        reshape_node_idx = model.get_node_id(new_node_names[j][0])
                        model.nodes[reshape_node_idx].attr = attr

                    elif concat_node_input_size == 3:
                        # the first new node
                        reshape_node_idx = model.get_node_id(new_node_names[j][0])
                        reshape_node = model.nodes[reshape_node_idx]

                        matmul_node_name = model.nodes[reshape_node_idx].output_tensors[0].dest_op[0]
                        matmul_node = model.get_node_by_name(matmul_node_name)
                        if matmul_node.op_type == 'MatMulWithBias':
                            # the 3D transpose_node: reshape2D -> Innerprodcut -> 3D transpose -> reshape
                            # reshape output to 2D for Innerprodcut
                            weight = matmul_node.input_tensors[1].data.shape[1]
                            attr = OrderedDict()
                            attr['dst_shape'] = '-1,' + str(weight)
                            reshape_node.input_tensors.pop(1)
                            model.nodes[reshape_node_idx].attr = attr

                            # restore to 3d after Innerprodcut
                            transpose_node = model.get_node_by_name(matmul_node.output_tensors[0].dest_op[0])
                            assert transpose_node.op_type == 'Transpose'

                            new_node_name = reshape_node.name + '_3D'
                            new_node_output_tensor_name = reshape_node.output_tensors[0].name + '_3D'

                            output_tensor = matmul_node.output_tensors[0].name + '_3D'
                            input_tensor = [matmul_node.output_tensors[0], model.nodes[0].output_tensors[0]]
                            output_tensor = [
                                Tensor(name=new_node_output_tensor_name,
                                       source_op=[new_node_name],
                                       dest_op=[transpose_node.name],
                                       dtype=reshape_node.output_tensors[0].dtype)
                            ]
                            new_node = util.construct_node(node_name=new_node_name,
                                                           op_type='Reshape',
                                                           input_tensors=input_tensor,
                                                           output_tensors=output_tensor)
                            attr = OrderedDict()
                            attr['dst_shape'] = '-1,-1,' + str(weight)
                            attr['dims'] = 0
                            new_node.attr = attr

                            # the place of insert is the next of the matmul node
                            matmul_node_idx = model.get_node_id(matmul_node.name)
                            transpose_node.input_tensors[0] = new_node.output_tensors[0]
                            model.insert_nodes(matmul_node_idx + 1, [new_node])
                        else:
                            matmul_node_name = ret_old_nodes[j][2].input_tensors[0].source_op[0]
                            matmul_node = model.get_node_by_name(
                                ret_old_nodes[j][2].input_tensors[0].source_op[0])
                            weight = matmul_node.input_tensors[1].data.shape[1]

                            attr = OrderedDict()
                            attr['dst_shape'] = '-1,-1,' + str(weight)
                            attr['dims'] = 0
                            model.nodes[reshape_node_idx].attr = attr

        return model



"""Regularizer."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .patterns import BasePattern
from intel_extension_for_transformers.optimization.utils.utility import LazyImport

torch = LazyImport('torch')

REGS = {}


def register_reg(name):
    """Register a regularizator to the registry.
    
    Args:
        name: A string that defines the scheduler type.

    Returns:
        cls: The class of register.
    """

    def register(reg):
        REGS[name] = reg
        return reg

    return register


def get_reg_type(config):
    """Obtain the regularizer type.
    
    Args:
        config: A config dict object that includes information of the regularizer.
    """
    for key in REGS.keys():  # pragma: no cover
        if config.get(key, None) != None:
            return key
    return None


def get_reg(config, modules, pattern):
    """Get registered regularizator class.
    
    Args:
        config: A config dict object that includes information of the regularizer.
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        pattern: A config dict object that includes information of the pattern.  
    """
    reg_type = config["reg_type"]
    if reg_type == None:
        return BaseReg(config, modules, pattern)
    if reg_type not in REGS.keys():  # pragma: no cover
        assert False, f"regularizator does not support {reg_type}, currently only support {REGS.keys()}"
    return REGS[reg_type](config, modules, pattern, config["reg_coeff"])


class BaseReg:
    """Regularizer.

    The class that performs regularization.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that includes information of the regularizer.
        pattern: A config dict object that includes information of the pattern.  
    """

    def __init__(self, config: dict, modules: dict, pattern: BasePattern):
        """Initialize."""
        self.modules = modules
        self.config = config
        self.pattern = pattern

    def on_before_optimizer_step(self):
        """Implement before optimizer.step()."""
        pass

    def on_after_optimizer_step(self):
        """Implement after optimizer.step()."""
        pass


@register_reg("group_lasso")
class GroupLasso(BaseReg):  # pragma: no cover
    """Regularizer.

    A regularizer class derived from BaseReg. In this class, the Group-lasso regularization will be performed.
    Group-lasso is a variable-selection and regularization method.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that includes information of the regularizer.
        pattern: A config dict object that includes information of the pattern.    

    Attributes:
        reg_terms: A dict {"module_name": Tensor} of regularization terms.
        alpha: A float representing the coeffient related to group lasso.
    """

    def __init__(self, config: dict, modules: dict, pattern: BasePattern, coeff):
        """Initialize."""
        super(GroupLasso, self).__init__(config, modules, pattern)
        assert "x" in self.config.pattern, "group lasso only supports NXM pattern"
        self.reg_terms = {}
        self.alpha = float(coeff)
        assert self.alpha >= 0, "group lasso only supports positive coeff"

    def on_before_optimizer_step(self):
        """Calculate the group-lasso score map."""
        with torch.no_grad():
            if self.pattern.invalid_layers == None:
                self.pattern.check_layer_validity()
            for key in self.modules.keys():
                if key in self.pattern.invalid_layers:
                    continue
                grad = self.modules[key].weight.grad
                reg_term = self.pattern.reshape_orig_to_pattern(grad, key)
                reg_term = self.alpha / (torch.norm(reg_term, p=2, dim=[1, 3]) + 1e-12)
                reg_term[torch.isinf(reg_term)] = 0.0
                self.reg_terms[key] = reg_term

    def on_after_optimizer_step(self):  ##decoupled with grad descent
        """Perform group lasso regularization after optimization."""
        with torch.no_grad():
            for key in self.modules.keys():
                if key in self.pattern.invalid_layers:
                    continue
                reg_term = self.pattern.reshape_reduced_to_orig(self.reg_terms[key], key,
                                                                self.modules[key].weight.shape)
                self.modules[key].weight -= reg_term * self.modules[key].weight



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The class of metric for optimization."""

class Metric(object):
    """Metric for optimization."""
    def __init__(self, name: str, greater_is_better: bool = True, is_relative: bool = True,
                 criterion: float = 0.01, weight_ratio: float = None):
        """Init an instance.

        Args:
            name: Metric name which evaluates function returns, like:"eval_f1", "eval_accuracy"...
            greater_is_better: Used to describe the usage of the metric, like: greater is better for f1, 
                this parameter is only used for quantization.
            is_relative: Used in conjunction with "criterion". If "criterion" is 0.01, and "is_relative" 
                is True, it means that we want to get an optimized model which metric drop <1% relative, 
                if "is_relative" is False, means metric drop <1% absolute, this parameter is only used 
                for quantization.
            criterion: Used in conjunction with "is_relative". If "criterion" is 0.01, and "is_relative" 
                is True, it means that we want to get an optimized model which metric drop <1% relative, 
                if "criterion" is 0.02, means metric drop <2% relative, this parameter is only used for 
                quantization.
            weight_ratio: Used when there are multiple metrics, for example: you want to focus on both 
                f1 and accuracy, then you will create f1 instance and accuracy instance, and indicate 
                their weight proportion. If weight_ratio of f1 is 0.3, and weight ratio of accuracy 
                is 0.7, then the final metric to tune is f1*0.3 + accuracy*0.7, this parameter is only 
                used for quantization.
        """
        self.name = name
        self.is_relative = is_relative
        self.criterion = criterion
        self.greater_is_better = greater_is_better
        self.weight_ratio = weight_ratio



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Transformer2Dmodel_FFNInputSlice_1 pattern."""

from .pattern import Pattern, pattern_registry
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='Transformer2Dmodel_FFNInputSlice_1')
class Transformer2Dmodel_FFNInputSlice_1(Pattern):
    """The Transformer2Dmodel_FFNInputSlice_1 pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_FFNInputSlice_1' graph.
    The search strategy is based on the following pattern mapping configs for the stable textEncoderV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'Transformer2Dmodel_FFNInputSlice_1': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Add'), (3, 'Div'), (4, 'Mul'),
                                (6, 'Slice')], [(3, 'Div'), (5, 'Mul'), (6, 'Slice')]],
                        'out': [[(0, 'Slice')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 6,
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            6: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1, 2, 3, 4, 5, 6]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['Transformer2Dmodel_FFNInputSlice_1'])):
            pattern_dict = pattern_mapping_config['Transformer2Dmodel_FFNInputSlice_1'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("Transformer2Dmodel_FFNInputSlice_1",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('Transformer2Dmodel_FFNInputSlice_1 mathched...')
                logger.debug('Transformer2Dmodel_FFNInputSlice_1 = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    # the first new node
                    assert ret_old_nodes[j][6].op_type == 'Slice'
                    slice_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[slice_node_idx].attr = ret_old_nodes[j][6].attr

                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
from ..graph_utils import names_from_input
from .. import logger


# graph.input
@operator_registry(operator_type='ONNXINPUT')
class ONNXINPUT(Operator):
    """Parse the ONNXINPUT operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def extract(self, framework, node, framework_model, nodes_dict, engine_graph=None):
        """Extract operators to the neural engine."""
        from ..onnx_utils import ONNX_DTYPE_ID
        self._name = node.name
        self._op_type = 'ONNXINPUT'
        output_tensor_name = names_from_input(self._name)[1]
        shape_len = len(node.type.tensor_type.shape.dim)
        shape = [-1] * shape_len
        for i, d in enumerate(node.type.tensor_type.shape.dim):
            param = d.dim_param
            if engine_graph and param in ['max_seq_len', 'seq_len'] and i == 0:
                engine_graph.add_config_item('seq_len_first_dim', True)
            if param == '':
                    v = d.dim_value
                    if v != 0:
                        shape[i] = v
                    else:
                        logger.error("Unknown dimension parameter in ONNX model input, " \
                                 "only dim_param or dim_value")
        dtype = ONNX_DTYPE_ID[node.type.tensor_type.elem_type]
        output_tensor = Tensor(
            name=output_tensor_name,
            shape=shape,
            dtype=dtype,
            source_op=[self._name],
            dest_op=nodes_dict[self._name].outputs,
        )

        self._output_tensors = [output_tensor]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The MergedEmbeddingbag pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='MergedEmbeddingbag')
class MergedEmbeddingbag(Pattern):
    """The MergedEmbeddingbag pattern.

    Fuse the original sub-graph into the custom acceleration 'MergedEmbeddingbag' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'MergedEmbeddingbag': [
                # DLRM onnx model
                {
                    'patterns': {
                        'in': [[(0, 'Split'), (1, 'Squeeze'), (2, 'Shape'),
                                (3, 'Gather'), (5, 'Unsqueeze'), (6, 'Concat'),
                                (7, 'Slice'), (8, 'Shape'), (9, 'Gather'), (10, 'Loop')],
                               [(), (4, 'Gather'), (6, 'Concat')]],
                    },
                },
            ]
        }

        pattern = pattern_mapping_config['MergedEmbeddingbag'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)

        if len(patterns_nodes_name) != 0:
            split_idx = model.get_node_id(patterns_nodes_name[0][0])
            split_node = model.nodes[split_idx]
            gather_idx = model.get_node_id(patterns_nodes_name[0][4])
            gather_node = model.nodes[gather_idx]
            match_nodes_name = []
            merged_embeddingbag_inputs = []
            merged_embeddingbag_outputs = []

            merged_embeddingbag_inputs.append(gather_node.input_tensors[0])
            merged_embeddingbag_inputs.append(split_node.input_tensors[0])

            for pattern_nodes_name in patterns_nodes_name:
                loop_idx = model.get_node_id(pattern_nodes_name[10])
                loop_node = model.nodes[loop_idx]
                weight = loop_node.input_tensors[0]
                merged_embeddingbag_inputs.append(weight)
                loop_output = loop_node.output_tensors[0]
                merged_embeddingbag_outputs.append(loop_output)
                for idx in range(len(pattern_nodes_name)-1):
                    match_nodes_name.append(pattern_nodes_name[idx])

            merged_embeddingbag_node_name = "MergedEmbeddingbag_" + split_node.name
            merged_embeddingbag_node = util.construct_node(merged_embeddingbag_node_name,
            'MergedEmbeddingbag', merged_embeddingbag_inputs, merged_embeddingbag_outputs)

            model.remove_nodes(match_nodes_name)
            model.insert_nodes(split_idx, [merged_embeddingbag_node])

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Optimization: provides the orchestrate optimizer for Pytorch."""
import logging
import os
import shlex

from neural_compressor.experimental import(
    common,
    Component,
    Distillation,
    Quantization,
    Pruning,
)
from neural_compressor.experimental.scheduler import Scheduler
from intel_extension_for_transformers.optimization import(
    DistillationConfig,
    Provider,
    QuantizationConfig,
    PruningConfig
)
from intel_extension_for_transformers.optimization.utils.utility import LazyImport
from intel_extension_for_transformers.optimization.quantization import QuantizationMode
from transformers import PreTrainedModel, PretrainedConfig
from transformers.file_utils import WEIGHTS_NAME
from typing import Callable, Optional, Union, List

torch = LazyImport("torch")

logger = logging.getLogger(__name__)


class Orchestrate_optimizer:
    """Orchestrate_optimizer aggregates and orchestrates components such as Quantization, Pruning and Distillation."""
    def __init__(
        self,
        model,
        components: Optional[List[Component]] = [],
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
        output_dir: Optional[str] = "saved_results",
    ):
        """Init an orchestrate optimizer.

        Args:
            model: Model to quantize and/or prune.
            components: List of Component objects which contains Quantization, Pruning, Distillation objects.
            eval_func: Evaluation function to evaluate the tuning objective.
            train_func: Training function which will be combined with pruning.
        """
        if len(components) == 0:
            raise RuntimeError("`NLPOptimizer` requires at least one `Quantization`, "
                               "`Pruning` or `Distillation` object")
        self.output_dir = output_dir
        self.model_config = model.config
        self.enable_inc_quant = False
        self.enable_inc_pruning = False
        self.scheduler = Scheduler()
        self.scheduler.model = common.Model(model)

        if len(components) > 1:
            agent = self.scheduler.combine(*components)
            agent.train_func = train_func
            agent.eval_func = eval_func
            for component in components:
                if isinstance(component, Distillation) and hasattr(component, 'criterion'):
                    agent.criterion = component.criterion
                if isinstance(component, Quantization):
                    self.enable_inc_quant = True
                if isinstance(component, Pruning):
                    self.enable_inc_pruning = True
            self.scheduler.append(agent)
        else:
            self.scheduler.append(*components)

    def fit(self):
        """Run the scheduler."""
        self.opt_model = self.scheduler()
        self.save_model(self.output_dir)
        if self.enable_inc_pruning == True:
            stats, sparsity = self.opt_model.report_sparsity()
            logger.info(stats)
            logger.info(sparsity)
        return self.opt_model.model

    def save_model(self, output_dir, tokenizer=None):
        """Save the model and tokenizer in the output directory.

        Args:
            output_dir: the path to save config.json and pytorch_model.bin.
            tokenizer (object, optional): the tokenizer object, use it if you want to 
                                          save tokenizer.json in output_dir. Defaults to None.
        """
        os.makedirs(shlex.quote(output_dir), exist_ok=True)
        torch.save(self.opt_model.quantized_state_dict(), os.path.join(shlex.quote(output_dir), WEIGHTS_NAME))
        if self.enable_inc_quant == True:
            self.model_config.torch_dtype = "int8"
        self.model_config.save_pretrained(output_dir)
        if tokenizer:   # pragma: no cover
            tokenizer.save_pretrained(output_dir)
        logger.info("orchestrate_optimizations model and configure file have saved to {}".format(
                    output_dir))


class NoTrainerOptimizer:   # pragma: no cover
    """Optimizer without using Trainer."""
    def __init__(
        self,
        model,
        output_dir: Optional[str] = "saved_results",
    ):
        """Init a NoTrainerOptimizer object.

        Args:
            model: FP32 model specified for low precision tuning.
            output_dir: The folder for saving the results.
        """
        self.model = model
        self.teacher_model = None
        self._eval_func = None
        self._train_func = None
        self._calib_func = None
        self._calib_dataloader = None
        self.output_dir = output_dir
        self.quant_config = None
        self.pruning_config = None
        self.distillation_config = None
        self._provider = Provider.INC.value
        self.pruner = None
        self.quantizer = None
        self.distiller = None
        self.in_training = False
        self.enable_inc_quant = False

    @property
    def eval_func(self):
        """Get the evaluation function."""
        return self._eval_func

    @property
    def train_func(self):
        """Get the train function."""
        return self._train_func

    @property
    def calib_func(self):
        """Get the calib function."""
        return self._calib_func

    @property
    def provider(self):
        """Get the provider."""
        return self._provider

    @property
    def calib_dataloader(self):
        """Get the calibration dataloader."""
        return self._calib_dataloader

    @eval_func.setter
    def eval_func(self, func: Callable):
        """Set the evaluation function.
            
        Args:
            func: evaluation function.
        """
        self._eval_func = func

    @train_func.setter
    def train_func(self, func: Callable):
        """Set the train function.
        
        Args:
            func: train function.
        """
        self._train_func = func

    @provider.setter
    def provider(self, provider):
        """Set the provider.
        
        Args:
            provider: optimization provider.
        """
        self._provider = provider

    @calib_dataloader.setter
    def calib_dataloader(self, dataloader):
        """Set the calibration dataloader.
        
        Args:
            dataloader: calibration dataloader.
        """
        self._calib_dataloader = dataloader

    def init_quantizer(
        self,
        quant_config,
        provider: str = Provider.INC.value,
    ):
        """Init a Quantization object with config.
        
        Args:
            quant_config: quantization config.
            provider: define the quantization provider.
        """
        from neural_compressor.experimental import Quantization

        assert isinstance(quant_config, QuantizationConfig), \
            "Please pass QuantizationConfig instance to trainer.quantize!"
        self.quant_config = quant_config
        self.metrics = self.quant_config.metrics
        self._provider = Provider[provider.upper()].value

        if self.quant_config.framework == "pytorch":
            if self.quant_config.approach == \
              QuantizationMode.POSTTRAININGDYNAMIC.value:
                self.quant_config.framework = "pytorch"
            else:
                self.quant_config.framework = "pytorch_fx"

        quantizer = Quantization(self.quant_config.inc_config)
        quantizer.model = common.Model(self.model)

        self.quantizer = quantizer
        return quantizer

    def _inc_quantize(
        self,
        quant_config,
        provider: str = Provider.INC.value,
    ):
        """Do the quantization."""
        if self.quantizer is None:
            self.init_quantizer(quant_config=quant_config, provider=provider)
        if self._eval_func is not None:
            self.quantizer.eval_func = self._eval_func
        if self._calib_func is not None:
            self.quantizer.calib_func = self._calib_func
        if self.quant_config.approach == QuantizationMode.POSTTRAININGSTATIC.value:
            assert self._calib_dataloader is not None, \
                "Please pass calib_dataloader to NoTrainerOptimizer.calib_dataloader"
            self.quantizer.calib_dataloader = self._calib_dataloader
        elif self.quant_config.approach == QuantizationMode.QUANTIZATIONAWARETRAINING.value:
            assert self._train_func is not None, \
                "Please pass train_func to NoTrainerOptimizer.train_func"
            self.quantizer.q_func = self._train_func
        self.opt_model = self.quantizer.fit()
        self.enable_inc_quant = True
        self.save_model(self.output_dir)
        return self.opt_model.model

    def quantize(
        self,
        quant_config: QuantizationConfig = None,
        provider: str = Provider.INC.value,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
        calib_func: Optional[Callable] = None,
        calib_dataloader=None,
    ):
        """Prepare for invoking the _inc_quantize function.
        
        Args:
            quant_config: quantization config.
            provider: define the quantization provider.
            eval_func: evaluation function.
            train_func: train function.
            calib_func: calibration function.
            calib_dataloader: calibration dataloader.
        """
        if eval_func is not None:
            self._eval_func = eval_func
        if train_func is not None:
            self._train_func = train_func
        if calib_func is not None:
            self._calib_func = calib_func
        if calib_dataloader is not None:
            self._calib_dataloader = calib_dataloader

        if self.quantizer is None:
            self._provider = Provider[provider.upper()].value

        if self._provider == Provider.INC.value:
            return self._inc_quantize(quant_config=quant_config, provider=provider)
        else:
            assert False, "Unsupport provider:{}".format(self._provider)

    def init_pruner(
        self,
        pruning_config = None,
        provider: str = Provider.INC.value,
    ):
        """Init a Pruning object with config.
        
        Args:
            pruning_config: pruning config.
            provider: define the pruning provider.
        """
        from neural_compressor.experimental import Pruning
        self.pruning_config = pruning_config
        self.metrics = self.pruning_config.metrics
        self._provider = Provider[provider.upper()].value

        assert isinstance(self.pruning_config, PruningConfig), \
            "please pass a instance of PruningConfig to trainer.prune!"

        pruner = Pruning(self.pruning_config.inc_config)
        pruner.model = common.Model(self.model)

        self.pruner = pruner
        return pruner

    def prune(
        self,
        pruning_config = None,
        provider: str = Provider.INC.value,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
    ):
        """Do the pruning.
        
        Args:
            pruning_config: pruning config.
            provider: define the pruning provider.
            eval_func: evaluation function.
            train_func: train function.
        """
        if self.pruner is None:
            self.init_pruner(pruning_config=pruning_config, provider=provider)
        if eval_func is not None:
            self._eval_func = eval_func
        if train_func is not None:
            self._train_func = train_func

        self.pruner.eval_func = self._eval_func

        self.pruner.pruning_func = self._train_func

        self.opt_model = self.pruner.fit()
        self.save_model(self.output_dir)
        stats, sparsity = self.opt_model.report_sparsity()
        logger.info(stats)
        logger.info(sparsity)

        return self.opt_model.model

    def init_distiller(
        self,
        distillation_config,
        teacher_model,
        provider: str = Provider.INC.value,
    ):
        """Init a Distillation object with config and the teacher model.
        
        Args:
            distillation_config: distillation config.
            teacher_model: set the teacher model.
            provider: define the distillation provider.
        """
        from neural_compressor.experimental import Distillation, common
        assert isinstance(distillation_config, DistillationConfig), \
            "please pass a instance of PruningConfig to trainer.prune!"
        self.distillation_config = distillation_config
        self._provider = Provider[provider.upper()].value
        self.metrics = self.distillation_config.metrics
        self.teacher_model = teacher_model

        distiller = Distillation(self.distillation_config.inc_config)
        distiller.model = common.Model(self.model)
        distiller.teacher_model = common.Model(self.teacher_model)

        self.distiller = distiller
        return distiller

    def distill(
        self,
        distillation_config,
        teacher_model,
        provider: str = Provider.INC.value,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
    ):
        """Do the distillation.
        
        Args:
            distillation_config: distillation config.
            teacher_model: set the teacher model.
            provider: define the distillation provider.
            eval_func: evaluation function.
            train_func: train function.
        """
        if self.distiller is None:
            self.init_distiller(
                distillation_config=distillation_config,
                teacher_model=teacher_model,
                provider=provider
            )
        if eval_func is not None:
            self._eval_func = eval_func
        if train_func is not None:
            self._train_func = train_func

        self.distiller.eval_func = self._eval_func
        self.distiller.train_func = self._train_func
        self.distiller.create_criterion()

        self.opt_model = self.distiller.fit()
        self.save_model(self.output_dir)
        return self.opt_model.model

    def _save_inc_int8(self, opt_model, output_dir):
        """Save the optimized model in the output directory.
        
        Args:
            opt_model: optimized model.
            output_dir: output path.
        """
        self.model.config.architectures = [self.model.__class__.__name__]
        self.model.config.torch_dtype = "int8"
        if isinstance(self.model.config, PretrainedConfig):
            self.model.config.save_pretrained(output_dir)
        weights_file = os.path.join(os.path.abspath(
          os.path.expanduser(output_dir)), WEIGHTS_NAME)
        torch.save(opt_model.quantized_state_dict(), weights_file)

    def save_model(self, output_dir, tokenizer=None):
        """Save the model and tokenizer in the output directory.

        Args:
            output_dir: the path to save config.json and pytorch_model.bin.
            tokenizer (object, optional): the tokenizer object, use it if you want to 
                                          save tokenizer.json in output_dir. Defaults to None.
        """
        os.makedirs(shlex.quote(output_dir), exist_ok=True)
        torch.save(self.opt_model.quantized_state_dict(), os.path.join(shlex.quote(output_dir), WEIGHTS_NAME))
        if self.enable_inc_quant and self.opt_model:
            self._save_inc_int8(self.opt_model, output_dir)
        else:
            self.model.save_pretrained(output_dir)
            self.model.config.save_pretrained(output_dir)
        if tokenizer:   # pragma: no cover
            tokenizer.save_pretrained(output_dir)
        logger.info("Optimized model and configure file have saved to {}".format(
                    output_dir))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The StableDiffusion_ReshapeFusion Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
from .. import logger


@pattern_registry(pattern_type='StableDiffusion_ReshapeFusion')
class StableDiffusion_ReshapeFusion(Pattern):
    """The StableDiffusion_ReshapeFusion pattern.

    Fuse the original sub-graph into the custom acceleration 'StableDiffusion_ReshapeFusion' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'StableDiffusion_ReshapeFusion': [
                {
                    # unet
                    'patterns': {
                        'in': [[(0, 'Gather'), (1, 'Reshape'), (2, 'Reshape'), (3, 'BinaryAdd'),
                                (4, 'Reshape'), (5, 'Reshape')]],
                        'out': [[(0, 'Gather')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            3: [0]
                        }, {
                            1: [1]
                        }], [[0, 1, 2, 3], 4]]
                    },
                    'output_tensors': {
                        0: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1, 2, 3]
                },
            ]
        }
        # fuse reshape nodes to previous innerproduct or matmul attr to enhance perf
        remove_node_name = []
        for node in model.nodes:
            if node.op_type == "Reshape" and node.input_tensors[0].source_op:
                pre_node = model.get_node_by_name(node.input_tensors[0].source_op[0])
                if pre_node.op_type in EXECUTOR_TYPE \
                   and (EXECUTOR_TYPE[pre_node.op_type] == "InnerProduct" or \
                        EXECUTOR_TYPE[pre_node.op_type] == "Matmul"):
                    pre_node.attr['reshape'] = node.attr['dst_shape']
                    pre_node.output_tensors[0] = node.output_tensors[0]
                    pre_node.output_tensors[0].source_op = [pre_node.name]
                    if node.output_tensors[0].dest_op != []:
                        next_node = model.get_node_by_name(node.output_tensors[0].dest_op[0])
                        for input_tensor in next_node.input_tensors:
                            if input_tensor.name == node.output_tensors[0].name:
                                input_tensor.source_op = [pre_node.name]
                    if 'dims' in node.attr:
                        pre_node.attr['reshape_dims'] = str(node.attr['dims'])
                        pre_node.input_tensors.append(node.input_tensors[1])

                    remove_node_name.append(node.name)

        model.remove_nodes(remove_node_name)

        for i in range(len(pattern_mapping_config['StableDiffusion_ReshapeFusion'])):
            pattern_dict = pattern_mapping_config['StableDiffusion_ReshapeFusion'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("StableDiffusion_ReshapeFusion",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('StableDiffusion_ReshapeFusion mathched...')
                logger.debug('StableDiffusion_ReshapeFusion = {}'.format(new_node_names))
                for i in range(len(new_node_names)):
                    offset = 0
                    if len(new_node_names[i]) == 3:
                        offset = 1
                    gather_node = model.get_node_by_name(new_node_names[i][offset])
                    attr = OrderedDict()
                    attr["axis"] = ret_old_nodes[i][0].attr['axis']
                    attr["batch_dims"] = ret_old_nodes[i][0].attr['batch_dims']
                    attr["append_op"] = "binary_add"
                    attr["reshape"] = ret_old_nodes[i][2].attr['dst_shape']
                    attr["reshape_dims"] = ret_old_nodes[i][2].attr['dims']
                    attr["mul"] = ret_old_nodes[i][2].attr['mul']
                    gather_node.attr = attr
                    if len(new_node_names[i]) >= 2:
                        import copy
                        newattr = copy.deepcopy(attr)
                        model.get_node_by_name(new_node_names[i][offset + 1]).attr = newattr
                    if len(ret_old_nodes[i][3].output_tensors[0].dest_op) == 3:
                        matMulWithBiasAdd_node = model.get_node_by_name(
                            ret_old_nodes[i][3].output_tensors[0].dest_op[1])
                        assert matMulWithBiasAdd_node.op_type == 'MatMulWithBiasAdd'
                        matMulWithBiasAdd_node.input_tensors[3] = gather_node.output_tensors[0]
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The AttentionBlock_Resize2Gather Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
import numpy as np
from .. import graph_utils as util
from ..ops import Tensor
from .. import logger


@pattern_registry(pattern_type='AttentionBlock_Resize2Gather')
class AttentionBlock_Resize2Gather(Pattern):
    """The AttentionBlock_Resize2Gather pattern.

    Fuse the original sub-graph into the custom acceleration 'AttentionBlock_Resize2Gather' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'AttentionBlock_Resize2Gather': [
                {
                    'patterns': {
                        'in': [[(0, 'Add'), (1, 'Div'), (2, 'Resize'), (3, 'Conv')]],
                        'out': [[(0, 'Add'), (1, 'Div'), (2, 'Gather'), (3, 'Gather'), (4, 'Conv')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1,
                        2: 'attentionBlock_Resize2Gather/extend_4th_dim',
                        3: 'attentionBlock_Resize2Gather/extend_3th_dim',
                        4: 3
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }], [[0, 1], 2]],
                        1: [[{
                            1: [1]
                        }], [[1], 2]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            3: [1]
                        }, {
                            3: [2]
                        }], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            1: [0]
                        }], [[0], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            3: [0]
                        }], [[0], 1]],
                    },
                    'returns': [0, 1, 2, 3]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['AttentionBlock_Resize2Gather'])):
            pattern_dict = pattern_mapping_config['AttentionBlock_Resize2Gather'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("AttentionBlock_Resize2Gather",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('AttentionBlock_Resize2Gather mathched...')
                logger.debug('AttentionBlock_Resize2Gather = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    # restor the add node
                    assert ret_old_nodes[j][0].op_type == 'Add'
                    add_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[add_node_idx].attr = ret_old_nodes[j][0].attr

                    # restore the Div node
                    assert ret_old_nodes[j][1].op_type == 'Div'
                    div_node_idx = model.get_node_id(new_node_names[j][1])
                    model.nodes[div_node_idx].attr = ret_old_nodes[j][1].attr

                    # restore the conv node
                    assert ret_old_nodes[j][3].op_type == 'Conv'
                    conv_node_idx = model.get_node_id(new_node_names[j][4])
                    model.nodes[conv_node_idx].attr = ret_old_nodes[j][3].attr

                    first_gather_node = model.get_node_by_name(new_node_names[j][2])
                    second_gather_node = model.get_node_by_name(new_node_names[j][3])

                    indices = (np.arange(16, dtype=np.int32) / 2).astype(np.int32)
                    extended_tensor_16 = Tensor(name='extend_range_to_16',
                                                source_op=[],
                                                dest_op=[],
                                                shape=[indices.shape[0]],
                                                data=indices,
                                                dtype='s32')

                    indices = (np.arange(128, dtype=np.int32) / 2).astype(np.int32)
                    extended_tensor_128 = Tensor(name='extend_range_to_128',
                                                 source_op=[],
                                                 dest_op=[],
                                                 shape=[indices.shape[0]],
                                                 data=indices,
                                                 dtype='s32')

                    indices = (np.arange(256, dtype=np.int32) / 2).astype(np.int32)
                    extended_tensor_256 = Tensor(name='extend_range_to_256',
                                                 source_op=[],
                                                 dest_op=[],
                                                 shape=[indices.shape[0]],
                                                 data=indices,
                                                 dtype='s32')

                    indices = (np.arange(512, dtype=np.int32) / 2).astype(np.int32)
                    extended_tensor_512 = Tensor(name='extend_range_to_512',
                                                 source_op=[],
                                                 dest_op=[],
                                                 shape=[indices.shape[0]],
                                                 data=indices,
                                                 dtype='s32')

                    extended_tensor_list = []
                    # hard code
                    if model.nodes[conv_node_idx].input_tensors[1].data.shape[0] == 1280:
                        extended_tensor_list.append(extended_tensor_16)
                    else:
                        extended_tensor_list.append(extended_tensor_128)
                    extended_tensor_list.append(extended_tensor_256)
                    extended_tensor_list.append(extended_tensor_512)

                    first_gather_node.input_tensors.insert(0, extended_tensor_list[j])
                    attr = OrderedDict()
                    attr['batch_dims'] = 3
                    attr['axis'] = 0
                    first_gather_node.attr = attr

                    second_gather_node.input_tensors.insert(0, extended_tensor_list[j])
                    attr = OrderedDict()
                    attr['batch_dims'] = 2
                    attr['axis'] = 0
                    second_gather_node.attr = attr

        return model



import copy
import torch
import numpy as np
import os
import shutil
import torch.utils.data as data
import unittest
from datasets import load_dataset, load_metric
from intel_extension_for_transformers.optimization import (
    PrunerConfig,
    PruningConfig,
    DistillationConfig,
    QuantizationConfig,
    DistillationCriterionMode,
    metrics,
    objectives,
    OptimizedModel,
)
from intel_extension_for_transformers.optimization.trainer import NLPTrainer
from intel_extension_for_transformers.optimization.distillation import Criterion

from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    set_seed,
)

os.environ["WANDB_DISABLED"] = "true"

class TestOrchestrateOptimizations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        set_seed(42)
        self.model = AutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased'
        )
        self.teacher_model = AutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased-finetuned-sst-2-english'
        )
        raw_datasets = load_dataset("glue", "sst2")["validation"]
        tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")
        def preprocess_function(examples):
            # Tokenize the texts
            args = (
                (examples['sentence'],)
            )
            result = tokenizer(*args, padding=True, max_length=64, truncation=True)
            return result
        raw_datasets = raw_datasets.map(
            preprocess_function, batched=True, load_from_cache_file=True
        )
        eval_dataset = raw_datasets.select(range(30))
        self.dataset = eval_dataset

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./orchestrate_optimizations_model', ignore_errors=True)

    def test_fx_orchestrate_optimization(self):
        metric = load_metric("accuracy")
        def compute_metrics(p):
            preds = p.predictions
            preds = np.argmax(preds, axis=1)
            return metric.compute(predictions=preds, references=p.label_ids)
        origin_weight = copy.deepcopy(self.model.classifier.weight)
        for mode in DistillationCriterionMode:
            print("Distillation approach:", mode.value)
            self.trainer = NLPTrainer(
                model=copy.deepcopy(self.model),
                train_dataset=self.dataset,
                eval_dataset=self.dataset,
                compute_metrics=compute_metrics,
            )
            self.trainer.calib_dataloader = self.trainer.get_eval_dataloader()
        tune_metric = metrics.Metric(
            name="eval_accuracy", is_relative=True, criterion=0.5
        )
        pruner_config = PrunerConfig(prune_type='PatternLock', target_sparsity_ratio=0.9)
        pruning_conf = PruningConfig(framework="pytorch_fx",pruner_config=[pruner_config], metrics=tune_metric)
        distillation_conf = DistillationConfig(framework="pytorch_fx", metrics=tune_metric)

        objective = objectives.performance
        quantization_conf = QuantizationConfig(
            approach="QuantizationAwareTraining",
            max_trials=600,
            metrics=[tune_metric],
            objectives=[objective]
        )

        from neural_compressor.adaptor.torch_utils.symbolic_trace import symbolic_trace
        self.model = symbolic_trace(self.model, is_qat=True)
        self.trainer.model = self.model
        conf_list = [pruning_conf, distillation_conf, quantization_conf]
        opt_model = self.trainer.orchestrate_optimizations(config_list=conf_list, teacher_model=self.teacher_model)
        self.assertTrue("quantize" in str(type(opt_model.classifier.module)))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ReshapeAfterRestoreHiddenStates Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import numpy as np
import copy


@pattern_registry(pattern_type='ReshapeAfterRestoreHiddenStates')
class ReshapeAfterRestoreHiddenStates(Pattern):
    """The ReshapeAfterRestoreHiddenStates pattern.

    Fuse the original sub-graph into the custom acceleration 'ReshapeAfterRestoreHiddenStates' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'ReshapeAfterRestoreHiddenStates': [
                # minilmv2-lat-roberta
                # reshape to 3d when do last scatter
                # and becomes to 2d when do inner product
               {
                    'patterns': {
                        'in': [[(0, 'ScatterElements'), (1, 'MatMulWithBias')]],
                        'out': [[(0, 'ScatterElements'), (1, 'Reshape'), (2, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 'reshape_to_2d_before_inner_product_in_last_restoration',
                        2: 1, 
                    },
                    'input_tensors': {
                        0: [[{0: [0]}, {0: [1]}, {0: [2]}], [[0, 1, 2], 3]],
                        1: [[], [[], 1]],
                        2: [[{1: [1]}, {1: [2]}], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{1: [0]}], [[0], 1]],
                    },
                    'returns': [0, 1]
                },
            ]
        }

        def _set_attr(se_attr, mat_attr, hidden_size, node_names, model):
            attr1 = OrderedDict()
            attr1['dst_shape'] = '-1,' + str(hidden_size)
            
            scatter_elements_node_idx = model.get_node_id(node_names[0])
            model.nodes[scatter_elements_node_idx].attr = se_attr
            reshape_2d_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_2d_node_idx].attr = attr1
            mat_node_idx = model.get_node_id(node_names[2])
            model.nodes[mat_node_idx].attr = mat_attr
        
        # minilmv2-lat-roberta
        pattern_dict = pattern_mapping_config['ReshapeAfterRestoreHiddenStates'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping(
                'ReshapeAfterRestoreHiddenStates', pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                hidden_size = int(ret_old_nodes[i][1].input_tensors[1].shape[0])
                se_attr = ret_old_nodes[i][0].attr
                mat_attr = ret_old_nodes[i][1].attr
                mat_node = model.get_node_by_name(new_node_names[i][2])
                reshape_node = model.get_node_by_name(new_node_names[i][1])
                mat_node.input_tensors[0].name = ret_old_nodes[i][1].input_tensors[0].name
                reshape_node.output_tensors[0].name =  mat_node.input_tensors[0].name
                _set_attr(se_attr, mat_attr, hidden_size, new_node_names[i], model)
            
            return model
        
        return model


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_MulReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_MulReshape')
class TextEncoder_MulReshape(Pattern):
    """The TextEncoder_MulReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_MulReshape' graph.
    This pattern is used for the reshape node after the QKV.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'TextEncoder_MulReshape': [
                # text encoder Reshape_3/4/5/7
                {
                    'patterns': {
                        'in': [[(0, 'Mul'), (1, 'Unsqueeze'), (2, 'Concat'), (3, 'Reshape')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3,
                    },
                    'input_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]],
                    },
                    'returns': [2, 0]
                },
            ]
        }

        def _set_attr(head_num, head_size, mul_size, node_names, model):
            attr = OrderedDict()
            attr['dst_shape'] = str(mul_size) + ',' + str(head_num) + ',' + str(head_size)

            reshape_node_idx = model.get_node_id(node_names[0])
            assert model.nodes[reshape_node_idx].op_type == 'Reshape'
            model.nodes[reshape_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['TextEncoder_MulReshape'])):
            pattern_dict = pattern_mapping_config['TextEncoder_MulReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_MulReshape",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('TextEncoder_MulReshape mathched...')
                logger.debug('TextEncoder_MulReshape = {}'.format(new_node_names))
                # except the reshape_7
                for j in range(len(new_node_names)):
                    pack_node = ret_old_nodes[j][0]
                    head_size = int(pack_node.input_tensors[-1].data)
                    head_num = int(pack_node.input_tensors[-2].data)
                    assert ret_old_nodes[j][1].op_type == 'Mul'
                    mul_size = int(ret_old_nodes[j][1].input_tensors[1].data)
                    _set_attr(head_num, head_size, mul_size, new_node_names[j], model)
                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine tensor file."""

from collections import OrderedDict
from .. import graph_utils as util
import numpy as np


class Tensor(object):
    """The definition of the neural engine."""
    def __init__(self,
                 name='',
                 source_op=None,
                 dest_op=None,
                 shape=None,
                 data=None,
                 dtype=None,
                 location=None):
        """The __init__ function."""
        self._name = name
        # assume data in tensor should be numpy array
        # however, we don't assign the data diretly if the tensor is
        # const like weight when parse model
        # otherwise it will make a bloated new graph
        # but it still can be set when using the constructed new graph
        self._data = data
        if shape is not None and len(shape) == 0:
            self._shape = [1]
        else:
            self._shape = shape
        self._dtype = dtype
        if not dtype and isinstance(data, np.ndarray):
            self._dtype = util.get_data_dtype(data)
        # location in bin file if const
        self._location = location
        if source_op == None:
            self._source_op = []
        else:
            self._source_op = source_op
        if dest_op == None:
            self._dest_op = []
        else:
            self._dest_op = dest_op

    @property
    def name(self):
        """Get the tensor name."""
        return self._name

    @name.setter
    def name(self, name):
        """Name Assignment."""
        self._name = name

    @property
    def data(self):
        """Get the tensor data."""
        return self._data

    @data.setter
    def data(self, data):
        """Data assignment."""
        self._data = data
        if data is not None:
            self._dtype = util.get_data_dtype(data)

    @property
    def shape(self):
        """Get the tensor shape."""
        return self._shape

    @shape.setter
    def shape(self, shape):
        """Shape assignment."""
        self._shape = shape

    @property
    def dtype(self):
        """Get the tensor dtype."""
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        """Dtype assignment."""
        self._dtype = dtype

    @property
    def location(self):
        """Get the tensor location."""
        return self._location

    @location.setter
    def location(self, location):
        """Location assignment."""
        self._location = location

    @property
    def source_op(self):
        """Get source_op."""
        return self._source_op

    @source_op.setter
    def source_op(self, source_op):
        """Source_op assignment."""
        self._source_op = source_op

    @property
    def dest_op(self):
        """Get dest_op."""
        return self._dest_op

    @dest_op.setter
    def dest_op(self, dest_op):
        """Dest_op assignment."""
        self._dest_op = dest_op

    @property
    def config(self):
        """Get the config dict in the graph."""
        conf_dict = OrderedDict()
        if self._dtype is not None:
            conf_dict['dtype'] = util.DTYPES_DICT.get(self._dtype, self._dtype)
        if self._shape is not None:
            conf_dict['shape'] = self._shape
        if self._location is not None:
            conf_dict['location'] = self._location

        return conf_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine subgraph matcher file."""

import copy
from tqdm import tqdm
from collections import namedtuple, OrderedDict
from .pattern import supported_patterns, superbert_patterns, PATTERNS
from .. import logger

EXECUTOR_TYPE = {
    "InnerProduct": "InnerProduct",
    "MatMulWithBias": "InnerProduct",
    "MatMulWithBiasAdd": "InnerProduct",
    "MatMulWithBiasGelu": "InnerProduct",
    "MatMulWithBiasTanh": "InnerProduct",
    "MatMulWithBiasRelu": "InnerProduct",
    "MatMulWithBiasSigmoid": "InnerProduct",
    "Matmul": "Matmul",
    "Einsum": "Matmul",
    "MatMul": "InnerProduct",
    "Conv": "Convolution",
    "QuantizedMatMulWithBiasAndDequantize": "InnerProduct",
    "TransposeBatchMatMul": "Matmul",
    "MatmulwithTranspose" : "Matmul", 
    "BatchMatMul": "Matmul",
    "BatchMatMulV2": "Matmul",
    "Add": "BinaryAdd",
    "AddV2": "BinaryAdd",
    "AddWithAdd": "BinaryAdd",
    "QLinearAdd": "BinaryAdd",
    "Transpose": "Reorder",
    "GatherV2": "Gather",
    "ExpandDimsToReshape": "Reshape",
    "QuantizeV2": "Quantize",
    "QuantizeLinear": "Quantize",
    "OneHot": "Onehot",
    "LayerNormalization": "LayerNorm",
    "FusedGemm": "InnerProduct",
    "_QuantizedFusedMatMulAndDequantize": "InnerProduct",
    "_FusedMatMul": "InnerProduct",
    "_MklLayerNorm": "LayerNorm",
    "Div": "BinaryOp",
    "Sub": "BinaryOp",
    "Mul": "BinaryOp",
    "Equal": "BinaryOp",
    "Less": "BinaryOp",
    "Greater": "BinaryOp",
    "LessEqual": "BinaryOp",
    "LessOrEqual": "BinaryOp",
    "GreaterOrEqual": "BinaryOp",
    "NonZero": "BinaryOp",
    "NotEqual": "BinaryOp",
    'Not': "BinaryOp",
    'Neg': "BinaryOp",
    "Sin": "CosSin",
    "Cos": "CosSin",
}

pattern_default_setting = {
    # General Pattern
    'PaddingSequence': True,
    'AttentionReshape': True,
    'QKVReshape': True,
    'ReshapeFusion': True,
    'InsertBF16Node': True,
    'OperatorAdaptor': True,

    'GroupNorm': True,

    # transpose_mode_int8
     'QKVMerge': False,

    # 'TextEncoder
    'TextEncoder_WordEmbedding': False,
    'TextEncoder_QReshape': False,
    'TextEncoder_KVReshape': False,
    'TextEncoder_AttentionMaskAddReshape': False,
    'TextEncoder_SoftmaxReshape': False,
    'TextEncoder_MulReshape': False,
    'TextEncoder_AttentionReshape': False,
    'TextEncoder_CasualAttentionMask': False,

    # vae deocder & Transformer2Dmodel
    'AttentionBlock_Resize2Gather': False,
    'AttentionBlock_QKVPreReshape': False,
    'AttentionBlock_AttentionMaskAddReshape': False,
    'AttentionBlock_ConstantOfShapeWithMul': False,

    'Transformer2Dmodel_GetSampleBatch': False,
    'Transformer2Dmodel_SampleSlice': False,
    'Transformer2Dmodel_EncoderHiddenStatesReshape': False,
    'Transformer2Dmodel_ConstantOfShapeWithMul': False,
    'Transformer2Dmodel_QKVPreReshape': False,
    'Transformer2Dmodel_QKVReshape': False,
    'AttentionBlock_QKVReshape': False,
    'Transformer2Dmodel_QKVReshapeTo4D': False,
    'Transformer2Dmodel_AttentionMaskAddReshape': False,
    'Transformer2Dmodel_FFNInputSlice': False,
    'Transformer2Dmodel_FFNInputSlice_1': False,
    'Transformer2DModel_UpBlockResize': False,

    # for all stable diffusion models
    'StableDiffusion_bf16Convert': False,
    'StableDiffusion_ReshapeFusion': False,
    
    #GPT-J
    'TorchEmbedding': True,
    'InnerproductReshapeFusion': True,
    'MatMulWithTranspose': True,
    'InnerproductWithBiasGelu': True,
    'SliceMask': True,
    'ArangewithReciprocal': True,
    'InnerproductwithSlice': True,
    'RoraryPosEmb': True,
    'EinsumwithArange': True,
    'RemoveSlice': True,
    'RemoveRange': True,
    'RemoveLastView': True,
    
    
    'TorchInsertBF16Node': True,
    'MultiHeadAttention': False,
    'Int8BF16MixedPrecisionChecker': False,
    'QuantizedGraphDtypeRefactor': True,
}

class SubGraphMatcher(object):
    """The SubGraphMatcher class."""
    def __call__(self, model, tune = False, pattern_config = None):
        """The __call__ function of SubGraphMatcher class."""
        logger.info('Start to implement Sub-Graph matching and replacing...') 
        if tune:
            model = self._tune_patterns(model)
        else:
            model = self._fuse_patterns(model, pattern_config=pattern_config)
        logger.info('Sub-Graph match and replace done...')
        return model

    def _fuse_patterns(self, model, supported_patterns=supported_patterns, pattern_mask=None, pattern_config=None):
        pattern_mask = [True for _ in range(len(supported_patterns))] \
                if pattern_mask == None else pattern_mask
        
        for index in range(len(supported_patterns)):
            pattern_name = supported_patterns[index]
            if pattern_name in pattern_default_setting:
                pattern_mask[index] = pattern_default_setting[pattern_name]

        # modify the pattern mask according to pattern_config
        if pattern_config != None:
            for index in range(len(supported_patterns)):
                pattern_name = supported_patterns[index]
                if pattern_name in pattern_config['pattern_switch']:
                    status = pattern_config['pattern_switch'][pattern_name]
                    pattern_mask[index] = status

        for pattern_id, pattern in enumerate(supported_patterns):
            if pattern in PATTERNS and pattern_mask[pattern_id]:
                p_fusion = PATTERNS[pattern]()
                model = p_fusion(model)
        model = self._remove_identity(model)
        return model

    def _tune_patterns(self, model, iterations = 10, warm_up = 5):
        # pattern tuning strategy(for superbert): 
        #    1. only one pattern off/on each time (pruning)
        #    2. check accuracy with framework
        #    3. and only save min latency config
        logger.info('Start tuning pattern...')
        all_patterns = supported_patterns + superbert_patterns
        pattern_mask = [True for i in range(len(all_patterns))]
        min_latency = float("inf")
        # skip tuning input node fusion and output node fusion
        for idx in tqdm(range(len(supported_patterns), len(all_patterns))):
            # pattern on
            on_latency = float("inf")
            try:
                on_model = copy.deepcopy(model)
                self._fuse_patterns(on_model, all_patterns, pattern_mask)
                on_result, on_latency = on_model._get_latency([], iterations, warm_up)
            except:
                logger.warning("Graph can not be inferenced, please check the graph!")
            # pattern off
            off_latency = float("inf")
            try:
                off_pattern_mask = copy.deepcopy(pattern_mask)
                off_pattern_mask[idx] = False
                off_model = copy.deepcopy(model)
                self._fuse_patterns(off_model, all_patterns, off_pattern_mask)
                off_result, off_latency = off_model._get_latency([], iterations, warm_up)
            except:
                logger.warning("Graph can not be inferenced, please check the graph!")
            # update min latency and pattern mask
            if off_latency < on_latency and off_latency < min_latency:
                min_latency = off_latency
                pattern_mask = off_pattern_mask
        
        # generate model according pattern mask 
        self._fuse_patterns(model, all_patterns, pattern_mask)
        logger.info('End tuning pattern...')
        return model

    def _remove_identity(self, model):
        rm_node_names = []
        rm_op_type = ['Identity']
        for i in range(len(model.nodes)):
            node = model.nodes[i]
            if node.op_type in rm_op_type:
                rm_node_names.append(node.name)
            else:
                if node.op_type in EXECUTOR_TYPE.keys():
                    if node.op_type == "Cos":
                        node.attr = OrderedDict({'algorithm': 'cos'})
                    if node.op_type == "Sin":
                        node.attr = OrderedDict({'algorithm': 'sin'})                           
                    op_type = EXECUTOR_TYPE[node.op_type]
                    model.nodes[i].op_type = op_type
        model.remove_nodes(rm_node_names)
        return model




import os
import shutil
import torch
import torch.utils.data as data
import unittest
from intel_extension_for_transformers.optimization import (
    AutoDistillationConfig,
    FlashDistillationConfig,
    metrics,
)
from intel_extension_for_transformers.optimization.trainer import NLPTrainer
from intel_extension_for_transformers.optimization.utils.utility import distributed_init
from transformers import (
    AutoModelForPreTraining,
    AutoTokenizer,
    TrainingArguments,
)

os.environ["WANDB_DISABLED"] = "true"


def main_worker(rank, world_size, model, teacher_model, dataset):
    try:
        distributed_init("gloo",
                         world_size=world_size,
                         rank=rank,
                         init_method='tcp://127.0.0.1:23456')
    except:
        distributed_init("gloo",
                         world_size=world_size,
                         rank=rank,
                         init_method='tcp://127.0.0.1:12345')

    training_args = TrainingArguments(
        output_dir='tmp_trainer',
        overwrite_output_dir=True,
        no_cuda=True,
        local_rank=rank
    )
    trainer = NLPTrainer(
            model=model,
            train_dataset=dataset,
            eval_dataset=dataset,
            args=training_args,
    )
    autodistillation_config = AutoDistillationConfig(
        search_space={'hidden_size': [128, 256], 'intermediate_size':[256, 512]},
        search_algorithm='Grid',
        metrics=[metrics.Metric(name="eval_loss", greater_is_better=False)],
        knowledge_transfer=FlashDistillationConfig(
            block_names=['bert.encoder.layer.0'],
            layer_mappings_for_knowledge_transfer=[
                [
                    [('bert.encoder.layer.0.output',)]
                ],
            ],
            train_steps=[3],
            loss_types=[['MSE']],),
        regular_distillation=FlashDistillationConfig(
            layer_mappings_for_knowledge_transfer=[
                [[('cls', '0')]]
            ],
            loss_types=[['KL']],
            add_origin_loss=[True],
            train_steps=[5]
        ),
    )
    best_model_archs = trainer.autodistillation(
        autodistillation_config,
        teacher_model,
        model_cls=AutoModelForPreTraining
    )
    assert len(best_model_archs) > 0, "Expected at least one best model archs."


class DummyDataset(data.Dataset):
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")
        self.sequence_a = "intel-extension-for-transformers is based in SH"
        self.sequence_b = "Where is intel-extension-for-transformers based? NYC or SH"
        self.encoded_dict = self.tokenizer(self.sequence_a, self.sequence_b)
        self.encoded_dict['labels'] = [-100] * len(self.encoded_dict['input_ids'])
        self.encoded_dict['labels'][1] = 17953
        self.encoded_dict['next_sentence_label'] = 0

    def __len__(self):
        return 1

    def __getitem__(self, index):
        """Returns one data pair (source and target)."""
        return self.encoded_dict


class TestAutoDistillation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForPreTraining.from_pretrained('prajjwal1/bert-tiny')
        self.teacher_model = AutoModelForPreTraining.from_pretrained('bert-base-uncased')
        self.dummy_dataset = DummyDataset()
        self.trainer = NLPTrainer(
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp_trainer', ignore_errors=True)

    def test_autodistillation(self):
        for search_algorithm in ['BO', 'Grid', 'Random']:
            max_trials = 6 if search_algorithm == 'Random' else 3
            autodistillation_config = \
              AutoDistillationConfig(
                search_space={'hidden_size': [128, 256], 'intermediate_size':[256, 512]},
                search_algorithm=search_algorithm,
                max_trials=max_trials,
                metrics=[metrics.Metric(name="eval_loss", greater_is_better=False)],
                knowledge_transfer=FlashDistillationConfig(
                    block_names=['bert.encoder.layer.0'],
                    layer_mappings_for_knowledge_transfer=[
                        [
                            [('bert.encoder.layer.0.output',)]
                        ],
                    ],
                    train_steps=[3],
                    loss_types=[['MSE']],),
                regular_distillation=FlashDistillationConfig(
                    layer_mappings_for_knowledge_transfer=[
                        [[('cls', '0')]]
                    ],
                    loss_types=[['KL']],
                    add_origin_loss=[True],
                    train_steps=[5]
                ),
            )
            best_model_archs = self.trainer.autodistillation(
                autodistillation_config,
                self.teacher_model,
                model_cls=AutoModelForPreTraining
            )
            # check best model architectures
            self.assertTrue(len(best_model_archs) > 0)

    def test_autodistillation_distributed(self):
        try:
            distributed_init() # for coverage purpose only
        except:
            pass
        ngpus_per_node = 2
        torch.multiprocessing.spawn(
            main_worker, nprocs=ngpus_per_node,
            args=(ngpus_per_node, self.model, self.teacher_model, self.dummy_dataset)
        )

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AutoDistillation: handling the whole pipeline of AutoDistillation for both pytorch and tensorflow framework."""

from pickletools import optimize
import numpy as np
import os
import random
import shutil
import tempfile

from functools import partial
from neural_compressor.conf.config import Conf, schema
from neural_compressor.conf.dotdict import DotDict
from neural_compressor.experimental import Distillation
from neural_compressor.experimental.nas.nas import NASBase
from neural_compressor.experimental.nas.nas_utils import find_pareto_front, NASMethods
from neural_compressor.experimental.nas.search_algorithms import \
    BayesianOptimizationSearcher, GridSearcher, RandomSearcher
from neural_compressor.strategy.bayesian import BayesianOptimization
from neural_compressor.utils import logger
from intel_extension_for_transformers.optimization.config import AutoDistillationConfig
from intel_extension_for_transformers.optimization.utils.utility import LazyImport

torch = LazyImport("torch")

class AutoDistillation(NASBase):
    """The framework class is designed for handling the whole pipeline of AutoDistillation.

    AutoDistillation is composed of three major stages, i.e. Model Exploration, Flash Distillation, 
    and Evaluation.
    In Model Exploration, a search engine will search for a better compressed model from the architecture 
    design space in each iteration.
    Flash Distillation is the stage for training the searched model to discover its potential.
    In Evaluation stage, the trained model will be evaluated to measure its performances (e.g. 
    the prediction accuracy, the hardware performance etc.) in order to select the best model architecture.
    """
    def __init__(self, model_builder, conf_fname_or_obj, framework='pytorch'):
        """Init an AutoDistillation instance base on config.

        Args:
        model_builder (function obj): A function to build model instance with the specified 
            model architecture parameters.
        conf_fname_or_obj (string or obj): The path to the YAML configuration file or
            a configuration object containing search setting, flash distillation settings, etc.
        framework: Specify the framework used as backend.
        """
        NASBase.__init__(self, search_space={}, model_builder=model_builder)
        self._train_func = None
        self._eval_func = None
        self.framework = framework.lower()
        self.init_by_cfg(conf_fname_or_obj)

    def search(self, res_save_path=None, model_cls=None):
        """Auto distillation search process.
        
        Returns:
            Best model architecture found in search process.
        """
        def reload_tf_model(model):
            with tempfile.TemporaryDirectory() as tmp_dir:
                model.save_pretrained(tmp_dir)
                assert model_cls, 'model_cls should not be None'
                model = model_cls.from_pretrained(tmp_dir)
            return model

        if res_save_path is None or not os.path.isdir(res_save_path):
            res_save_path = os.getcwd()
        save_path = os.path.join(res_save_path, 'AutoDistillationResults')
        self.model_paras_num = {}
        logger.info(f'search save_path = {save_path}')
        self.load_search_results(save_path)
        os.makedirs(save_path, exist_ok=True)

        for i in range(self.max_trials):
            logger.info(
                "{fix} Trial {n} starts, {r} trials to go {fix}".format(
                    n=i+1, r=self.max_trials-i-1, fix="="*30
                )
            )
            if self.framework == 'pytorch':
                if not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0:
                    model_arch_paras = self.select_model_arch()
                    logger.info("Model architecture {} proposed.".format(model_arch_paras))
                if torch.distributed.is_initialized():
                    model_arch_paras_sync = [model_arch_paras] \
                        if torch.distributed.get_rank() == 0 else [None]
                    torch.distributed.broadcast_object_list(model_arch_paras_sync, src=0)
                    model_arch_paras = model_arch_paras_sync[0]
            else:
                model_arch_paras = self.select_model_arch()
            model = self._model_builder(model_arch_paras)
            if self.framework == 'tensorflow':
                model = reload_tf_model(model)

            model_paras = self.count_model_parameters(model)
            logger.info(
                "***** Number of model parameters: {:.2f}M *****".format(model_paras / 10**6)
            )
            self.model_paras_num[tuple(model_arch_paras.values())] = model_paras
            if tuple(model_arch_paras.values()) in self.search_results:
                logger.info("Skip evaluated model architecture {}.".format(model_arch_paras))
                continue
            if tuple(model_arch_paras.values()) in self.resumed_search_results:
                logger.info(
                    "Find previous results of model architecture: {}.".format(model_arch_paras)
                )
                metrics = self.resumed_search_results[tuple(model_arch_paras.values())]
            else:
                logger.info("Assessing model architecture: {}.".format(model_arch_paras))
                metrics = self.estimate(model)
            logger.info(
                "Metrics of model architecture {} is {}.".format(model_arch_paras, metrics)
            )
            self.search_results[tuple(model_arch_paras.values())] = metrics

            if (self.framework != "pytorch" or not torch.distributed.is_initialized() 
              or torch.distributed.get_rank() == 0):
                self._search_algorithm.get_feedback(sum(self.metrics_conversion(metrics)))
                print(f'res_save_path: {res_save_path}, save_path = {save_path}')
                os.makedirs(save_path, exist_ok=True)
                self.dump_search_results(
                    os.path.join(save_path, 'Trial_{}_results.txt'.format(i + 1)))

        if self.framework != "pytorch" or not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0:
            for model_arch_vec in self.resumed_search_results:
                if model_arch_vec not in self.search_results:
                    self.search_results[model_arch_vec] = \
                        self.resumed_search_results[model_arch_vec]
                    model = self._model_builder(self.params_vec2params_dict(model_arch_vec))
                    if self.framework == 'tensorflow':
                        model = reload_tf_model(model)
                    self.model_paras_num[model_arch_vec] = self.count_model_parameters(model)
            self.dump_search_results(os.path.join(save_path, 'Final_results.txt'.format(i+1)))
        self.find_best_model_archs()
        logger.info(
            "{fix} Found {n} best model architectures {fix}".format(
                n=len(self.best_model_archs), fix="="*30
            )
        )
        for i, model_arch in enumerate(self.best_model_archs):
            logger.info("Best model architecture {}: {}".format(i+1, model_arch))
        return self.best_model_archs

    def estimate(self, model):
        """Train and evaluate the model.

        Returns:
            Evaluated metrics of the model.
        """
        assert self._train_func is not None and self._eval_func is not None, \
            "train_func and eval_func must be set."
        model = self._train_func(model)
        return self._eval_func(model)

    def count_model_parameters(self, model):
        if self.framework == 'pytorch':
            if isinstance(model, torch.nn.Module):
                return sum(p.numel() for p in model.parameters())
            else: # pragma: no cover
                raise NotImplementedError("Only support torch model now.")
        elif self.framework == 'tensorflow':
            return model.num_parameters()

    def load_search_results(self, path):
        """Load previous search results.
        
        Args:
            path: The file path which stores the previous results.
        """
        self.resumed_search_results = {}
        lastest_results_record = os.path.join(path, 'lastest_results.npy')
        if not os.path.exists(path) or not os.path.exists(lastest_results_record):
            return
        self.resumed_search_results = np.load(lastest_results_record, allow_pickle=True).item()
        if self.framework == "pytorch" and torch.distributed.is_initialized():
            torch.distributed.barrier()
        if self.framework != "pytorch" or not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0:
            os.makedirs(os.path.join(path, 'previous_results'), exist_ok=True)
            for f in os.listdir(path):
                if os.path.isfile(os.path.join(path, f)):
                    shutil.move(os.path.join(path, f), os.path.join(path, 'previous_results', f))
        logger.info("Loaded previous results.")

    def init_by_cfg(self, conf_fname_or_obj):
        """Use auto distillation config to init the instance of autodistillation."""
        if isinstance(conf_fname_or_obj, str): # pragma: no cover
            if os.path.isfile(conf_fname_or_obj):
                self.config = Conf(conf_fname_or_obj).usr_cfg
            else:
                raise FileNotFoundError(
                    "{} is not a file, please provide a file path.".format(conf_fname_or_obj)
                )
        elif isinstance(conf_fname_or_obj, AutoDistillationConfig):
            self.config = conf_fname_or_obj.config
            if self.framework == 'pytorch':
                schema.validate(self.config)
        else: # pragma: no cover
            raise NotImplementedError(
                "Please provide a str path to config file or a config dict."
            )
        assert self.config.auto_distillation is not None, "auto_distillation section must be set"
        auto_distillation_cfg = self.config.auto_distillation
        # search related
        self.init_search_cfg(auto_distillation_cfg)
        # flash distillation related
        self.flash_distillation_config = auto_distillation_cfg.flash_distillation

    def create_distillers(self):
        """Create flash and regular distillers."""
        def create_tf_distiller(distillation_cfg):
            block_names = distillation_cfg.block_names \
                if distillation_cfg.block_names else ['']
            train_steps = distillation_cfg.train_steps \
                if distillation_cfg.train_steps else [500]
            conf = DotDict({
                'model':{
                    'name': 'flash_distillation_{}'.format(0), 'framework': self.framework
                },
                'distillation':{
                    'train':{'optimizer': {'SGD':{'learning_rate': 1e-4}},
                            'criterion': {'KnowledgeDistillationLoss':
                                        {'temperature': distillation_cfg.temperature,
                                        'loss_types': distillation_cfg.loss_types,
                                        'loss_weights': distillation_cfg.loss_weights,
                                        }}}
                }
            })
            return [Distillation(conf)], block_names, train_steps

        def create_distiller(distillation_cfg):
            distillers = []
            assert distillation_cfg.layer_mappings_for_knowledge_transfer, \
                'layer_mappings_for_knowledge_transfer should be defined in flash_distillation'
            layer_mappings = \
                distillation_cfg.layer_mappings_for_knowledge_transfer
            block_names = distillation_cfg.block_names \
                if distillation_cfg.block_names else [''] * len(layer_mappings)
            loss_types = distillation_cfg.loss_types \
                if distillation_cfg.loss_types else [['MSE']*len(_) for _ in layer_mappings]
            loss_weights = distillation_cfg.loss_weights \
                if distillation_cfg.loss_weights else [[1.0 / len(_)] * len(_) \
                    for _ in layer_mappings]
            add_origin_loss = distillation_cfg.add_origin_loss \
                if distillation_cfg.add_origin_loss else [False] * len(layer_mappings)
            train_steps = distillation_cfg.train_steps \
                if distillation_cfg.train_steps else [500] * len(layer_mappings)
            temperatures = distillation_cfg.temperatures \
                if distillation_cfg.temperatures else [1.0] * len(layer_mappings)
            assert len(layer_mappings) == len(block_names) == len(loss_types) == \
                len(loss_weights) == len(add_origin_loss) == len(train_steps), \
                "lengths of layer_mappings_for_knowledge_transfer, block_names, " + \
                "loss_types, loss_weights, add_origin_loss and train_steps should be the same."
            distillers = []
            for i, lm in enumerate(layer_mappings):
                conf = DotDict({
                    'model':{
                        'name': 'flash_distillation_{}'.format(i), 'framework': self.framework
                    },
                    'distillation':{
                        'train':{'optimizer': {'SGD':{'learning_rate': 1e-4}},
                                'criterion': {'IntermediateLayersKnowledgeDistillationLoss':
                                            {'layer_mappings': lm,
                                                'loss_types': loss_types[i],
                                                'loss_weights': loss_weights[i],
                                                'add_origin_loss': add_origin_loss[i]}}}
                    }
                })
                distillers.append(Distillation(conf))
            return distillers, block_names, train_steps

        flash_distillation_config = self.flash_distillation_config
        # knowledge transfer related
        self.flash_distillers = []
        if flash_distillation_config.knowledge_transfer:
            knowledge_transfer_cfg = flash_distillation_config.knowledge_transfer
            if self.framework == 'pytorch':
                self.flash_distillers, self.flash_block_names, self.flash_train_steps = \
                    create_distiller(knowledge_transfer_cfg)
            else:
                self.flash_distillers, self.flash_block_names, self.flash_train_steps = \
                    create_tf_distiller(knowledge_transfer_cfg)

        # regular distillation related
        self.regular_distillers = []
        if flash_distillation_config.regular_distillation:
            regular_distillation_cfg = flash_distillation_config.regular_distillation
            if self.framework == 'pytorch':
                self.regular_distillers, self.regular_block_names, self.regular_train_steps = \
                    create_distiller(regular_distillation_cfg)
            else:
                self.regular_distillers, self.regular_block_names, self.regular_train_steps = \
                    create_tf_distiller(regular_distillation_cfg)

    @property
    def teacher_model(self):
        """Getter of teacher model."""
        return self._teacher_model

    @teacher_model.setter
    def teacher_model(self, user_model):
        """Setter of teacher model."""
        self._teacher_model = user_model

    @property
    def student_model(self):
        """Getter of student model."""
        return self._model

    @student_model.setter
    def student_model(self, user_model):
        """Setter of student model."""
        self._model = user_model

    @property
    def train_func(self):
        """Getter of train function."""
        return self._train_func

    @train_func.setter
    def train_func(self, train_func):
        """Setter of train function."""
        self._train_func = train_func

    @property
    def eval_func(self):
        """Getter of evaluation function."""
        return self._eval_func

    @eval_func.setter
    def eval_func(self, eval_func):
        """Setter of evaluation function."""
        self._eval_func = eval_func

    def __repr__(self): # pragma: no cover
        """Return class name."""
        return 'AutoDistillation'


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The loader file."""

from ..graph_utils import LazyImport, get_model_fwk_name
import os
from .. import logger

onnx = LazyImport('onnx')
tf = LazyImport('tensorflow')
onnxoptimizer = LazyImport('onnxoptimizer')
torch = LazyImport('torch')

class Loader(object):
    """Load the model into the frontend of different inference frameworks."""
    def __call__(self, model, pattern_config=None):
        """Chceck if the model is the tensorflow or onnxruntime."""
        framework = get_model_fwk_name(model)
        """Extract the node attr from tensorflow."""
        if framework == 'tensorflow':
            if isinstance(model, str):
                graph = tf.Graph()
                graph_def = tf.compat.v1.GraphDef()
                with open(model, 'rb') as f:
                    graph_def.ParseFromString(f.read())
                    with graph.as_default():
                        tf.import_graph_def(graph_def, name='')
                config = tf.compat.v1.ConfigProto()
                model = tf.compat.v1.Session(graph=graph, config=config)
        """Extract the node attr from onnxruntime."""
        if framework == 'onnxruntime':
            if isinstance(model, str):
                model = onnx.load(model)
                
                try:
                    from ..onnx_utils import ONNX_OPTIMIZER_PASS
                    optimize_level = os.getenv('ONNX_OPTIMIZER_LEVEL', 1)
                    passes = [v for k, v in ONNX_OPTIMIZER_PASS.items() if k <= optimize_level]
                    # for usage, see: https://github.com/onnx/optimizer/blob/master/onnxoptimizer/
                    # __init__.py#L25
                    model = onnxoptimizer.optimize(model, passes, fixed_point=False)
                    onnx.save(model, "optmodel.onnx")
                    logger.info("Try to optimize onnx model use onnxoptimizer and "\
                                    "optimize passes are {}".format(passes))
                except BaseException:
                    pass
        if framework == 'torch':
            if isinstance(model, str):
                model = torch.jit.load(model)
                model = torch.jit.freeze(model.eval())
        return model, framework



import copy
import os
import shutil
import torch.utils.data as data
import unittest
from intel_extension_for_transformers.optimization import (
    metrics,
    OptimizedModel,
    PrunerConfig,
    PruningConfig,
    PruningMode,
    NoTrainerOptimizer
)
from intel_extension_for_transformers.optimization.trainer import NLPTrainer
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer
)

os.environ["WANDB_DISABLED"] = "true"



class DummyDataset(data.Dataset):
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")
        self.sequence_a = "intel-extension-for-transformers is based in SH"
        self.sequence_b = "Where is intel-extension-for-transformers based? NYC or SH"
        self.encoded_dict = self.tokenizer(self.sequence_a, self.sequence_b)
        self.encoded_dict['labels'] = 1

    def __len__(self):
        return 1

    def __getitem__(self, index):
        """Returns one data pair (source and target)."""
        return self.encoded_dict


class TestPruning(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased'
        )
        self.dummy_dataset = DummyDataset()
        self.trainer = NLPTrainer(
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )
        self.optimizer = NoTrainerOptimizer(self.model)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./pruned_model', ignore_errors=True)

    def test_fx_model_prune(self):
        origin_weight = copy.deepcopy(self.model.classifier.weight)
        for mode in PruningMode:
            # not supported yet
            if mode.name != "BasicMagnitude".upper():
                continue
            self.trainer = NLPTrainer(
                model=self.model,
                train_dataset=self.dummy_dataset,
                eval_dataset=self.dummy_dataset,
            )
            metric = metrics.Metric(name="eval_loss")
            pruner_config = PrunerConfig(prune_type=mode.name, target_sparsity_ratio=0.9)
            pruning_conf = PruningConfig(pruner_config=pruner_config, metrics=metric)
            agent = self.trainer.init_pruner(pruning_config=pruning_conf)
            pruned_model = self.trainer.prune()
            # By default, model will be saved in tmp_trainer dir.
            self.trainer.save_model('./pruned_model')
            loaded_model = OptimizedModel.from_pretrained(
                './pruned_model',
            )
            pruned_weight = copy.deepcopy(pruned_model.classifier.weight)
            loaded_weight = copy.deepcopy(loaded_model.classifier.weight)
            # check pruned model
            self.assertTrue((pruned_weight != origin_weight).any())
            # check loaded model
            self.assertTrue((pruned_weight == loaded_weight).all())

    def test_functional_prune(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        self.trainer = NLPTrainer(self.model)
        pruner_conf = PrunerConfig(prune_type='BasicMagnitude', target_sparsity_ratio=0.9)
        pruning_conf = PruningConfig(pruner_config=pruner_conf)
        self.trainer.prune(pruning_conf,
                           provider="inc",
                           train_func = train_func,
                           eval_func = eval_func,)

    def test_no_trainer_prune(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        pruner_conf = PrunerConfig(prune_type='BasicMagnitude', target_sparsity_ratio=0.9)
        pruning_conf = PruningConfig(pruner_config=pruner_conf)
        self.optimizer.eval_func = eval_func
        self.optimizer.train_func = train_func
        self.optimizer.prune(pruning_conf,
                           provider="inc",
                           train_func = train_func,
                           eval_func = eval_func,)
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The RemoveLastView Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
import numpy as np
import copy

@pattern_registry(pattern_type='RemoveLastView')
class RemoveLastView(Pattern):
    """The RemoveLastView pattern.

    Fuse the original sub-graph into the custom acceleration 'RemoveLastView' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        remove_list = []
        for node in model.nodes:
            if node.op_type == 'View':
                pre_node = model.get_node_by_name(node.input_tensors[0].source_op[0])
                dst_node = model.get_node_by_name(node.output_tensors[0].dest_op[0])
                node.output_tensors[0].source_op = [pre_node.name]
                pre_node.output_tensors[0] = node.output_tensors[0]
                dst_node.input_tensors[0].source_op = [pre_node.name]
                for in_tensor in node.input_tensors[1:]:
                    source_node = model.get_node_by_name(in_tensor.source_op[0])
                    if len(source_node.output_tensors) == 1 and len(source_node.output_tensors[0].dest_op) == 1 \
                        and source_node.output_tensors[0].dest_op[0] == node.name:
                        remove_list.append((source_node.name))
                remove_list.append(node.name)
        model.remove_nodes(remove_list)
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for data augmentation."""

import nlpaug.augmenter.char as nac
import nlpaug.augmenter.sentence as nas
import nlpaug.augmenter.word as naw
from enum import Enum


EOS = '</s>'


class AugmenterType(Enum):
    """Enumeration of types of augmentation."""
    TEXTGENERATIONAUG = "textgenerationaug"
    KEYBOARDAUG = "KeyboardAug"
    OCRAUG = "OcrAug"
    SPELLINGAUG = "SpellingAug"
    CONTEXTUALWORDEMBSFORSENTENCEAUG = "ContextualWordEmbsForSentenceAug"


AUGMENTER_MAPPING = {
    AugmenterType.KEYBOARDAUG.value: nac,
    AugmenterType.OCRAUG.value: nac,
    AugmenterType.SPELLINGAUG.value: naw,
    AugmenterType.CONTEXTUALWORDEMBSFORSENTENCEAUG.value: nas,

}


def get_augmenter_from_type(aug_type: str):
    """Get nlpaug's augmenter by augment_type name.

    The nlpaug is a library helps you with augmenting nlp for your machine learning projects.
    It provide many augmenter, please refer to https://github.com/makcedward/nlpaug#augmenter.
    """
    assert aug_type in AUGMENTER_MAPPING, "Unspported the augmenter type:{}".format(aug_type)
    return AUGMENTER_MAPPING[aug_type]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine tensorflow utils."""

import numpy as np
import re
from collections import namedtuple, OrderedDict
from . import logger
from .ops.tensor import Tensor
from . import graph_utils as util


def create_tf_node(op, name, inputs):
    """Create a nodedef object.

    Args:
        op (string): op type
        name (string): op name
        inputs (string list): op's inputs name

    Returns:
        nodedef: the created nodedef object
    """
    from tensorflow.core.framework import node_def_pb2

    new_node = node_def_pb2.NodeDef()
    new_node.op = op
    new_node.name = name
    for input_name in inputs:
        new_node.input.extend([input_name])
    return new_node


def graph_node_names_details(nodes):
    """Parse the graph nodes ans get the graph_nodes_dict.

    Be used for Grpah class when converting a tensorflow computation graph to an engine graph.
    The node_name is the key, node in value is for getting the Const
    tensor value and the input_tensor source op; outputs in value is for
    output_tensor dest op.
    Args:
        nodes (tendorflow graph_def.node): NodeDef list

    Returns:
        node_names_details: the graph node info dict

    """
    # nodes = model.graph_def.node
    # some node may have several output edges
    # use tensor represents edge in graph
    node_details = namedtuple('node_details', ['node', 'outputs'])
    node_names_details = {}
    for node in nodes:
        node_name = node.name
        each_node = node_details(node=node, outputs=[])
        if node_name not in node_names_details:
            node_names_details[node_name] = each_node
    for name, details in node_names_details.items():
        # node.input represents input tensors name, not nodes names
        # in many times, they are same string list
        inputs = []
        for each_input in details.node.input:
            each_input = util.names_from_input(each_input)[0]
            if each_input not in inputs:
                inputs.append(each_input)
        for each_i in inputs:
            node_names_details[each_i].outputs.append(name)
    return node_names_details


def get_tensor_dest_op(node_name, tensor_name, nodes_dict):
    """Get the tensor dest op name.

    Args:
        node_name: string, the source node name of tensor
        tensor_name: string, with ':0' or something like it
        nodes_dict: dict returned by the graph_node_names_details function

    Returns:
        dest_op_names: list, store the tensor's dest op names
    """
    dest_op_names = []
    post_node_names = nodes_dict[node_name].outputs
    for n_name in post_node_names:
        for input_name in nodes_dict[n_name].node.input:
            if util.names_from_input(input_name)[1] == tensor_name:
                dest_op_names.append(n_name)
            else:
                continue
    return dest_op_names


def tf_extract_operator(node, framework_model, nodes_dict, engine_graph=None):
    """Decorate the operator in tensorflow.

    Args:
        node: NodeDef
        framework_model: TensorflowModel
        nodes_dict: dict, return value from graph_node_names_details
        tf_dtypes: dict, for get the dtype string
        engine_graph: Engine Graph class

    Returns:
        op_type: node op type
        input_tensors: Tensor list, contains the node input tensors info
        output_tensors: Tensor list, contains the node output tensor info
    """
    from tensorflow.python.framework import tensor_util

    op_type = node.op
    input_tensors = []
    output_tensors = []

    input_names = node.input
    for i, n in enumerate(input_names):
        # for Identity Constant
        if n.endswith('/read'):
            input_names[i] = n[:-5]
        else:
            continue

    """ input_tensors
    each input_tensor has its own soure op, but same dest op
    so both have single string
    """
    # name list
    input_tensor_names = input_names
    for input_tensor_name in input_tensor_names:
        input_node_name, input_tensor_name = util.names_from_input(input_tensor_name)
        pre_node = nodes_dict[input_node_name].node
        if pre_node.op == 'Const':
            data = tensor_util.MakeNdarray(pre_node.attr['value'].tensor)
            dtype = util.get_data_dtype(data)
            shape = list(data.shape) if data.shape != () else [1]
            input_tensor = Tensor(name=input_tensor_name,
                                  source_op=[],
                                  dest_op=[node.name],
                                  shape=shape,
                                  data=data,
                                  dtype=dtype
                                  )
            input_tensors.append(input_tensor)
        else:
            input_tensor = Tensor(name=input_tensor_name,
                                  source_op=[input_node_name],
                                  dest_op=[node.name],
                                  shape=None,
                                  data=None,
                                  dtype=None
                                  )
            input_tensors.append(input_tensor)

    """ output_tensors
    Almost every op generate one tensor in deep learning, so we just give one tensor in 
    output_tensors (list). However, this tensor maybe delivered to several nodes, so the 
    dest_op should have several strings.
    """
    if node.op not in MULTI_OUTPUT_OP.keys():
        out_num = 1
    else:
        if node.op == 'Unpack':
            out_num = node.attr['num'].i
        elif node.op == 'IteratorGetNext':
            out_num = len(node.attr['output_types'].list.type)
        else:
            out_num = MULTI_OUTPUT_OP[node.op]

    for i in range(out_num):
        output_tensor_name = node.name + ':' + str(i)
        try:
            output_tensor = framework_model.graph.get_tensor_by_name(output_tensor_name)
        except BaseException:
            return op_type, input_tensors, []

        dest_op = get_tensor_dest_op(node.name, output_tensor_name, nodes_dict)
        output_tensor = Tensor(name=output_tensor_name,
                               source_op=[node.name],
                               dest_op=dest_op,
                               shape=None,
                               data=None,
                               dtype=None
                               )

        output_tensors.append(output_tensor)

    return op_type, input_tensors, output_tensors

"""Get tensor dtype info when meets its encoded value in a tensorflow node"""
TF_DTYPE_ID = {3: 'int32',
               1: 'fp32',
               9: 'int64',
               7: 'string', }

"""Deal with the nodes which have multiple indeterminate outputs in tf_extract_operator function"""
MULTI_OUTPUT_OP = {'Unpack': -1,
                   'IteratorGetNext': -1,
                   }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine onnx extractor file.

   Extract all nodes in the input onnx model and convert them to engine python operators.
   All these python operators will compose the engine graph in the order of the original onnx
   calculation graph. Please noticed that the mapping between operators is not one-to-one
   correspondence. For more related deatils, please refer to 'ops' APIs.
"""

from .. import logger
from ..graph.graph import Graph
from ..ops.op import OPERATORS
from ..onnx_utils import graph_node_names_details
from ..graph_utils import names_from_input
from intel_extension_for_transformers.backends.neural_engine.compile.ops.tensor import Tensor


class ONNXExtractor(object):
    """The ONNXExtractor class.

    Decorate the node in model.graph_def, and the new node has the attributes like input_tensors
    and output_tensors, these tensors record the source/dest op name. All of these nodes 
    (in a list) will compose a graph, which is Graph class, as the return object.

    Args:
        model: ONNXModel

    Return:
        Graph: Graph class, the new graph object
    """
    @classmethod
    def __call__(self, model):
        """The __call__ function of the extractor."""
        # onnx_consts = model.initializer()
        graph_nodes_dict = graph_node_names_details(model)
        logger.info('Start to extarct onnx model ops...')
        new_graph = Graph()
        new_graph.framework_modeling_config['framework'] = 'onnxruntime'
        input_tensors_name = [names_from_input(n.name)[1] for n in model.graph.input]
        output_tensors_name = [names_from_input(n.name)[1] for n in model.graph.output]
        new_graph.input_tensors_name = input_tensors_name
        new_graph.output_tensors_name = output_tensors_name
        for graph_input in model.graph.input:
            op_type = 'ONNXINPUT'
            new_node = OPERATORS[op_type]()
            new_node.extract('onnxruntime', graph_input, model, graph_nodes_dict, new_graph)
            new_graph.insert_nodes(len(new_graph.nodes), [new_node])
        for node in model.graph.node:
            op_type = node.op_type
            if op_type == 'Constant':
                continue
            else:
                if op_type not in OPERATORS.keys():
                    op_type = "OpAny"
                new_node = OPERATORS[op_type]()
                new_node.extract('onnxruntime', node, model, graph_nodes_dict)
                new_graph.insert_nodes(len(new_graph.nodes), [new_node])
                if op_type == 'Loop':
                    for inner_node in node.attribute[0].g.node:
                        op_type = inner_node.op_type
                        if op_type == 'Constant':
                            continue
                        else:
                            import intel_extension_for_transformers.backends.neural_engine.compile.graph_utils\
                                as util
                            input_tensor_names = inner_node.input
                            for input_tensor_name in input_tensor_names:
                                origin_tensor_name, input_tensor_name = \
                                    util.names_from_input(input_tensor_name)
                                if origin_tensor_name in graph_nodes_dict and \
                                    hasattr(graph_nodes_dict[origin_tensor_name], 'node'):
                                    pre_node = graph_nodes_dict[origin_tensor_name].node
                                    data = None
                                    graph_node = new_graph.get_node_by_name(node.name)
                                    has_tensor = True
                                    for tensor in graph_node.input_tensors:
                                        if origin_tensor_name + ':0' == tensor.name:
                                            has_tensor = False
                                    if pre_node in model.graph.initializer and has_tensor:
                                        from onnx.numpy_helper import to_array
                                        data = to_array(pre_node)
                                        shape = list(data.shape) if data.shape != () else [1]
                                        dtype = util.get_data_dtype(data)
                                        loop_tensor = Tensor(name=origin_tensor_name,
                                                                source_op=[],
                                                                dest_op=[node.name],
                                                                shape=shape,
                                                                data=data,
                                                                dtype=dtype
                                                            )
                                        new_graph.change_node_input_tensors(node.name, 0, \
                                                                     loop_tensor, mode='insert')

        return new_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor


# The inputs must be two-dimensional matrices
@operator_registry(operator_type='_QuantizedFusedMatMulAndDequantize')
class _QuantizedFusedMatMulAndDequantize(Operator):
    """Parse the _QuantizedFusedMatMulAndDequantize operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):

        # for baremetal, if QuantizedMatMulWithBiasAndDequantize has post op
        # its output_dtype maybe the origin dtype in attributes, float32
        # but if it has not, use int8 (s8) for performance consideration
        """Extract the node attr from tensorflow."""
        if framework == 'tensorflow':
            self._attr['output_dtype'] = 'fp32'  # 's8'
            transpose_a = node.attr['transpose_a'].b
            transpose_b = node.attr['transpose_b'].b
            if transpose_a:
                self._attr['src0_perm'] = '1,0'
            if not transpose_b:
                self._attr['src1_perm'] = '1,0'

            epsilon = node.attr['epsilon'].f
            if epsilon != 0.0:
                self._attr['epsilon'] = epsilon

            # if 'leakyrelu_alpha' in attributes.attr.keys():
            #     leakyrelu_alpha = attributes.attr['leakyrelu_alpha'].f
            # else:
            #     leakyrelu_alpha = None
            # if leakyrelu_alpha != None:
            #     self.attrs.attr['append_op'] = 'relu'
            #     self.attrs.attr['leakyrelu_alpha'] = leakyrelu_alpha
            fused_ops = node.attr['fused_ops'].list.s
            fused_ops = [str(i, encoding='utf-8') for i in fused_ops]
            self._attr['fused_ops'] = fused_ops



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine base extractor file.

   This is the base extractor class for tensorflow and onnx.
   Please refer to onnx_extractor.py and tf_extractor.py for more details.
"""

from .tf_extractor import TensorflowExtractor
from .onnx_extractor import ONNXExtractor
from .torch_extractor import TorchExtractor
from .. import logger
from ..graph_utils import get_model_fwk_name

EXTRACTORS = {
    'tensorflow': TensorflowExtractor,
    'onnxruntime': ONNXExtractor,
    'torch': TorchExtractor,
}


class Extractor(object):
    """Extractor base class.

    A super class for an operation extractor.
    Do additional extraction of operation attributes without modifying of graph topology.
    """

    def __call__(self, model, pattern_config = None):
        """The __call__ funtion of the base extractor class."""
        framework = model[1]
        extractor = EXTRACTORS[framework]()
        model = extractor(model[0])
        logger.info('Extract {} model done...'.format(framework))
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Transformer2Dmodel_SampleSlice pattern."""

from .pattern import Pattern, pattern_registry
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='Transformer2Dmodel_SampleSlice')
class Transformer2Dmodel_SampleSlice(Pattern):
    """The Transformer2Dmodel_SampleSlice pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_SampleSlice' graph.
    The search strategy is based on the following pattern mapping configs for the stable textEncoderV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'Transformer2Dmodel_SampleSlice': [
                {
                    'patterns': {
                        'in': [[(0, 'Concat'), (1, 'Slice'), (3, 'Concat'), (4, 'Cast')],
                               [(0, 'Concat'), (2, 'Slice'), (3, 'Concat')]],
                    },
                },
            ]
        }

        pattern = pattern_mapping_config['Transformer2Dmodel_SampleSlice'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        logger.info('Transformer2Dmodel_SampleSlice mathched...')
        logger.debug('Transformer2Dmodel_SampleSlice = {}'.format(patterns_nodes_name))
        if len(patterns_nodes_name) != 0:
            for j in range(len(patterns_nodes_name)):
                #the first node
                concat_node = model.get_node_by_name(patterns_nodes_name[j][0])
                assert concat_node.op_type == 'Concat'
                if concat_node.attr['axis'] == -1:
                    concat_node.attr['axis'] = 1

                # the second node
                slice_node = model.get_node_by_name(patterns_nodes_name[j][1])
                assert slice_node.op_type == 'Slice'
                if slice_node.attr['ends'] == '-1':
                    slice_node.attr['ends'] = '320'

                #the third node
                concat_node = model.get_node_by_name(patterns_nodes_name[j][3])
                assert concat_node.op_type == 'Concat'
                if concat_node.attr['axis'] == -1:
                    concat_node.attr['axis'] = 1

                cast_node = model.get_node_by_name(patterns_nodes_name[j][4])
                assert cast_node.op_type == 'Cast'
                gemm_node = model.get_node_by_name(cast_node.output_tensors[0].dest_op[0])
                assert gemm_node.op_type == 'MatMulWithBias'

                concat_node.output_tensors[0].dest_op = gemm_node.name
                gemm_node.input_tensors[0] = concat_node.output_tensors[0]
                model.remove_nodes([cast_node.name])

            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The CollectQuantInfo Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
import numpy as np
import copy

@pattern_registry(pattern_type='CollectQuantInfo')
class CollectQuantInfo(Pattern):
    """The CollectQuantInfo pattern.

    Fuse the original sub-graph into the custom acceleration 'CollectQuantInfo' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'CollectQDQInfo': [
                {
                    'patterns': {
                        'in': [[(0, 'Quantize'), (1, ['DequantizeLinear'])]],
                    },
                },
                {
                    'patterns': {
                        'in': [[(0, 'DequantizeLinear')]],
                    },
                },
            ],
            'CollectTorchInfo': [
                {
                    'patterns': {
                        'in': [[(0, 'Quantize'), (1, ['Dequantize'])]],
                    },
                    
                },
                {
                    'patterns': {
                        'in': [[(0, 'Dequantize')]],
                    },
                },
            ],
            'RemoveQuantDequant': [
                {
                    'patterns': {
                        'in': [[(0, 'Quantize')]]
                    }
                },
                {
                    'patterns': {
                        'in': [[(0, 'DequantizeLinear')]]
                    }
                }
            ],
            'QLinearMatMul': [
                {
                    'patterns': {
                        'in': [[(0, 'QLinearMatMul')]],
                        'out': [[(0, 'MatMul')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [3]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                }
            ]
        }

        def CollectQDQInfo(model):
            # Collect the activation quant info
            pattern = pattern_mapping_config['CollectQDQInfo'][0]['patterns']['in']
            patterns_nodes_name = util.search_pattern(pattern, model)
            util.quant_info_init()
            new_dict = {}
            for pattern_nodes_name in patterns_nodes_name:
                quant_node = model.get_node_by_name(pattern_nodes_name[0])
                dquant_node = model.get_node_by_name(pattern_nodes_name[1])
                dquant_output = dquant_node.output_tensors[0]
                dtype = "s8" if dquant_node.input_tensors[2].data.dtype == 'int8' else "u8"
                max_range = 127 if dtype == "s8" else 255
                quant_max = (max_range -
                             dquant_node.input_tensors[2].data) * dquant_node.input_tensors[1].data
                quant_min = quant_max - 255 * dquant_node.input_tensors[1].data

                dtype = dtype + "_insert"
                util.insert_quant_info(quant_node.input_tensors[0].name, [quant_min, quant_max, dtype])
                for dst_op in dquant_output.dest_op:
                    dst_node = model.get_node_by_name(dst_op)
                    for idx, input_tensor in enumerate(dst_node.input_tensors):
                        if input_tensor.name == dquant_output.name:
                            for pre_quant_name in quant_node.input_tensors[0].source_op:
                                pre_quant_node = model.get_node_by_name(pre_quant_name)
                                if pre_quant_node.op_type in EXECUTOR_TYPE and \
                                    (EXECUTOR_TYPE[pre_quant_node.op_type] == "InnerProduct" or \
                                    EXECUTOR_TYPE[pre_quant_node.op_type] == "Matmul"):
                                    quant_info = util.get_quant_info()
                                    output_dtype = "output_" + dtype
                                    if model.get_node_by_name(pre_quant_node.input_tensors[0].
                                                              source_op[0]).op_type == "Transpose":
                                        input_tensor_name = model.get_node_by_name(
                                            pre_quant_node.input_tensors[0].source_op[0]
                                        ).input_tensors[0].name
                                        origin_quant_min = quant_info[
                                            pre_quant_node.input_tensors[0].name][0]
                                        origin_quant_max = quant_info[
                                            pre_quant_node.input_tensors[0].name][1]
                                        util.insert_quant_info(input_tensor_name, [
                                            origin_quant_min, origin_quant_max, output_dtype,
                                            quant_min, quant_max
                                        ])
                                    elif model.get_node_by_name(pre_quant_node.input_tensors[0].
                                                                source_op[0]).op_type == "Softmax":
                                        input_tensor_name = pre_quant_node.input_tensors[0].name
                                        origin_quant_min = quant_info[
                                            pre_quant_node.input_tensors[0].name][0]
                                        origin_quant_max = quant_info[
                                            pre_quant_node.input_tensors[0].name][1]
                                        util.insert_quant_info(input_tensor_name, [
                                            origin_quant_min, origin_quant_max, output_dtype,
                                            quant_min, quant_max
                                        ])
                                    else:
                                        input_tensor_name = pre_quant_node.input_tensors[1].name
                                        new_dict[input_tensor_name] = [quant_min, quant_max]

                                if pre_quant_node.op_type == "Transpose":
                                    util.insert_quant_info(pre_quant_node.input_tensors[0].name,
                                                           [quant_min, quant_max, dtype])
                                pre_quant_node.output_tensors[0].dest_op.append(dst_node.name)

                                dst_node.input_tensors[idx] = pre_quant_node.output_tensors[0]

                model.remove_nodes([pattern_nodes_name[0], pattern_nodes_name[1]])

            # Collect the weight quant info
            pattern = pattern_mapping_config['CollectQDQInfo'][1]['patterns']['in']
            patterns_nodes_name = util.search_pattern(pattern, model)
            for pattern_nodes_name in patterns_nodes_name:
                dquant_node = model.get_node_by_name(pattern_nodes_name[0])
                dquant_output = dquant_node.output_tensors[0]
                dtype = "s8" if dquant_node.input_tensors[2].data.dtype == 'int8' else "u8"
                dtype = dtype + "_weight"
                max_range = 127 if "s8" in dtype else 255
                quant_max = (max_range -
                             dquant_node.input_tensors[2].data) * dquant_node.input_tensors[1].data
                quant_min = quant_max - 255 * dquant_node.input_tensors[1].data
                util.insert_quant_info(dquant_node.input_tensors[0].name,
                                       [quant_min, quant_max, dtype])
                if dquant_output.name in new_dict:
                    old_quant_min = new_dict[dquant_output.name][0]
                    old_quant_max = new_dict[dquant_output.name][1]
                    dtype = "output_" + dtype
                    util.insert_quant_info(dquant_node.input_tensors[0].name,
                                           [quant_min, quant_max, dtype, old_quant_min, old_quant_max])
                for dst_op in dquant_output.dest_op:
                    matmul_node = model.get_node_by_name(dst_op)
                    # fall back int8 weight matmul op_type
                    if matmul_node.op_type == "BatchMatMul":
                        matmul_node.op_type = "MatMul"
                    for idx, input_tensor in enumerate(matmul_node.input_tensors):
                        if input_tensor.name == dquant_output.name:
                            matmul_node.input_tensors[idx] = dquant_node.input_tensors[0]
                model.remove_nodes([pattern_nodes_name[0]])

        def RemoveQuantDequant(model):
            # Remove quant/dequant
            remove_nodes_list = []
            for qpattern_info in pattern_mapping_config['RemoveQuantDequant']:
                qpattern = qpattern_info['patterns']['in']
                qnodes_name = util.search_pattern(qpattern, model)
                for qnode_name in qnodes_name:
                    qnode = model.get_node_by_name(qnode_name[0])
                    for qnode_input in qnode.input_tensors:
                        if not isinstance(qnode_input.data, np.ndarray):
                            # connect quant/dequant pre node with next node
                            qpre_node_name = qnode_input.source_op[0]
                            qpre_node = model.get_node_by_name(qpre_node_name)
                            qpre_node_out_dest_op = copy.deepcopy(qpre_node.output_tensors[0].dest_op)
                            # 1. the dest of previous node's output tensor connects to next node
                            for qpre_node_dst_op in qpre_node_out_dest_op:
                                if qpre_node_dst_op == qnode.name:
                                    qpre_node.output_tensors[0].dest_op.remove(qnode.name)
                                    qnode_output = qnode.output_tensors[0]
                                    for qnext_node in qnode_output.dest_op:
                                        qpre_node.output_tensors[0].dest_op.append(qnext_node)
                            # 2. the input of next node is current node's input
                            # 3. the source node of next node's input is previous node
                            qnode_output = qnode.output_tensors[0]
                            for qnext_node_name in qnode_output.dest_op:
                                qnext_node = model.get_node_by_name(qnext_node_name)
                                for idx in range(len(qnext_node.input_tensors)):
                                    if qnext_node.input_tensors[idx].name == qnode_output.name:
                                        qnext_node.input_tensors[idx] = qnode_input
                                        qnext_node.input_tensors[idx].source_op = [qpre_node.name]
                    remove_nodes_list.append(qnode_name[0])
            model.remove_nodes(remove_nodes_list)

        def QLinearMatMul(model):
            def scale2minmax(dtype, scale, zp):
                maxq = 127 if dtype == "s8" else 255
                quant_max = (maxq - zp) * scale
                quant_min = quant_max - 255 * scale
                return quant_min, quant_max

            def set_attr(new_node_names, model):
                for i in range(len(new_node_names)):
                    mat_node_idx = model.get_node_id(new_node_names[i][0])
                    attr = OrderedDict()
                    # see OneDNN InnerProduct related requirements
                    attr['transpose_a'] = False
                    attr['transpose_b'] = False
                    model.nodes[mat_node_idx].attr = attr
            # Collect the quant info
            pattern_dict = pattern_mapping_config['QLinearMatMul'][0]
            model, matmul_nodes_name, qmatmul_nodes = \
                        util.pattern_mapping("QLinearMatMul", pattern_dict, model)
            util.quant_info_init()
            for qmatmul_node in qmatmul_nodes:
                qmatmul_node = qmatmul_node[0]
                # input
                input = qmatmul_node.input_tensors[0]
                input_scale = qmatmul_node.input_tensors[1]
                input_zp = qmatmul_node.input_tensors[2]
                # input min max info
                input_dtype = "s8" if input_zp.data.dtype == "int8" else "u8"
                input_qdtype = input_dtype + "_weight" \
                            if type(input.data) == np.ndarray else input_dtype + "_insert"
                input_min, input_max = scale2minmax(input_dtype, input_scale.data, input_zp.data)
                util.insert_quant_info(input.name, [input_min, input_max, input_qdtype])
                # weight
                weight = qmatmul_node.input_tensors[3]
                weight_scale = qmatmul_node.input_tensors[4]
                weight_zp = qmatmul_node.input_tensors[5]
                # weight min max info
                weight_dtype = "s8" if weight_zp.data.dtype == "int8" else "u8"
                weight_qdtype = weight_dtype + "_weight" \
                            if type(weight.data) == np.ndarray else weight_dtype + "_insert"
                weight_min, weight_max = scale2minmax(weight_dtype, weight_scale.data, weight_zp.data)
                # util.insert_quant_info(weight.name, [weight_min, weight_max, weight_qdtype])
                # weight_s8 -> weight_fp32
                if type(weight.data) == np.ndarray :
                    weight.data = (weight.data-weight_zp.data) * weight_scale.data
                # output
                output = qmatmul_node.output_tensors[0]
                output_scale = qmatmul_node.input_tensors[6]
                output_zp = qmatmul_node.input_tensors[7]
                # output min max info
                output_dtype = "s8" if output_zp.data.dtype == "int8" else "u8"
                output_qdtype = "output_" + weight_qdtype # trick for insert_quant
                output_min, output_max = scale2minmax(output_dtype, output_scale.data, output_zp.data)
                # trick: for insert_quant_node.py
                util.insert_quant_info(weight.name, \
                        [weight_min, weight_max, output_qdtype, output_min, output_max])
                # trick: for insert_quant_node.py
                for id in range(len(qmatmul_node.input_tensors)):
                    for pre_op_name in qmatmul_node.input_tensors[id].source_op:
                        pre_op = model.get_node_by_name(pre_op_name)
                        pre_out_tensor_name = pre_op.output_tensors[0].name
                        qmatmul_in_tensor_name = qmatmul_node.input_tensors[id].name
                        if pre_out_tensor_name == qmatmul_in_tensor_name:
                            # trick: collect softmax output min/max(fp32 in, int8 out in neural engine)
                            if pre_op.op_type == "Softmax":
                                origin_scale = qmatmul_node.input_tensors[id+1]
                                origin_zp = qmatmul_node.input_tensors[id+2]
                                origin_min, origin_max = scale2minmax(origin_zp.data.dtype,
                                                            origin_scale.data, origin_zp.data)
                                output_scale = qmatmul_node.input_tensors[-2]
                                output_zp = qmatmul_node.input_tensors[-1]
                                output_min, output_max = scale2minmax(output_zp.data.dtype,
                                                            output_scale.data, output_zp.data)
                                output_zp_dtype = "s8" if output_zp.data.dtype == 'int8' else "u8"
                                output_dtype = "output_" + output_zp_dtype + "_insert"
                                util.insert_quant_info(pre_out_tensor_name,
                                    [origin_min, origin_max, output_dtype, output_min, output_max])
                            # trick: pre transpose has the same min/max as matmul,
                            elif pre_op.op_type == "Transpose":
                                qmatmul_in_scale = qmatmul_node.input_tensors[id+1]
                                qmatmul_in_zp = qmatmul_node.input_tensors[id+2]
                                qmatmul_in_min, qmatmul_in_max = scale2minmax(qmatmul_in_zp.data.dtype,
                                                            qmatmul_in_scale.data, qmatmul_in_zp.data)
                                trans_in_tensor_name = pre_op.input_tensors[0].name
                                qmatmul_in_zp_dtype = "s8" if qmatmul_in_zp.data.dtype == 'int8' else "u8"
                                qmatmul_in_dtype = qmatmul_in_zp_dtype + "_insert"
                                util.insert_quant_info(trans_in_tensor_name,
                                    [qmatmul_in_min, qmatmul_in_max, qmatmul_in_dtype])
                # matmul has tranpose attr, but qlinear matmul has not
                set_attr(matmul_nodes_name, model)

        def CollectTorchInfo(model):
            # Collect the activation quant info
            pattern = pattern_mapping_config['CollectTorchInfo'][0]['patterns']['in']
            patterns_nodes_name = util.search_pattern(pattern, model)
            new_dict = {}
            for pattern_nodes_name in patterns_nodes_name:
                quant_node = model.get_node_by_name(pattern_nodes_name[0])
                dquant_node = model.get_node_by_name(pattern_nodes_name[1])
                scale = quant_node.attr['scale']
                zp = quant_node.attr['zero_point']
                dquant_output = dquant_node.output_tensors[0]
                dtype = "u8" if quant_node.attr['dtype'] == 13 else "s8"
                max_range = 127 if dtype == "s8" else 255
                quant_max = np.array((max_range - zp) * scale)
                quant_min = np.array(quant_max - 255 * scale)

                dtype = dtype + "_insert"
                map_name = quant_node.input_tensors[0].name
                pre_node = quant_node
                while True:
                    pre_node = model.get_node_by_name(pre_node.input_tensors[0].source_op[0])
                    if pre_node.op_type == 'Reorder':
                        map_name = pre_node.input_tensors[0].name
                    else:
                        break
                util.insert_quant_info(map_name, [quant_min, quant_max, dtype])
                for dst_op in dquant_output.dest_op:
                    dst_node = model.get_node_by_name(dst_op)
                    for idx, input_tensor in enumerate(dst_node.input_tensors):
                        if input_tensor.name == dquant_output.name:
                            for pre_quant_name in quant_node.input_tensors[0].source_op:
                                pre_quant_node = model.get_node_by_name(pre_quant_name)
                                for pre_quant_node_out_tensor in pre_quant_node.output_tensors:
                                    if pre_quant_node_out_tensor.name == quant_node.input_tensors[0].name:
                                        if quant_node.name in pre_quant_node_out_tensor.dest_op:
                                            pre_quant_node_out_tensor.dest_op.remove(quant_node.name)
                                        pre_quant_node_out_tensor.dest_op.append(dst_node.name)
                                        dst_node.input_tensors[idx] = pre_quant_node_out_tensor
                model.remove_nodes([pattern_nodes_name[0], pattern_nodes_name[1]])
            # Collect the weight quant info
            pattern = pattern_mapping_config['CollectTorchInfo'][1]['patterns']['in']
            patterns_nodes_name = util.search_pattern(pattern, model)
            for pattern_nodes_name in patterns_nodes_name:
                dquant_node = model.get_node_by_name(pattern_nodes_name[0])
                dquant_output = dquant_node.output_tensors[0]
                for dst_op in dquant_output.dest_op:
                    dst_node = model.get_node_by_name(dst_op)
                    for idx, input_tensor in enumerate(dst_node.input_tensors):
                        if input_tensor.name == dquant_output.name:
                            dquant_node.input_tensors[0].dest_op.remove(dquant_node.name)
                            dquant_node.input_tensors[0].dest_op.append(dst_node.name)
                            dst_node.input_tensors[idx] = dquant_node.input_tensors[0]
                model.remove_nodes([pattern_nodes_name[0]])

        if model.framework_modeling_config['framework'] == 'torch':
            CollectTorchInfo(model)
            return model

        # if ONNX QDQ model, only CollectQDQInfo
        # if ONNX QLinear model, RemoveQuantDequant and QLinearMatMul
        is_qlinear_graph = False
        for node in model.nodes:
            if node.op_type == "QLinearMatMul":
                is_qlinear_graph = True
        if not is_qlinear_graph:
            CollectQDQInfo(model)
        else:
            RemoveQuantDequant(model)
            QLinearMatMul(model)

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor


# tf.cast(x, dtype, name=None)
# input dtype -> dest stype
@operator_registry(operator_type='Cast')
class Cast(Operator):
    """Parse the Cast operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from tensorflow."""
        if framework == 'tensorflow':
            from ..tf_utils import TF_DTYPE_ID
            self._attr['SrcT'] = TF_DTYPE_ID[node.attr['SrcT'].type]
            self._attr['DstT'] = TF_DTYPE_ID[node.attr['DstT'].type]
            self._attr['Truncate'] = node.attr['Truncate'].b
        """Extract the node attr from onnxruntime."""
        if framework == 'onnxruntime':
            from ..onnx_utils import ONNX_DTYPE_ID
            self._attr['DstT'] = ONNX_DTYPE_ID[node.attribute[0].i]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for pytorch framework."""

import importlib
import os
from neural_compressor.utils.utility import LazyImport

def distributed_init(backend="gloo", world_size=1, rank=-1, init_method=None,
                     master_addr='127.0.0.1', master_port='12345'):
    """Init the distibute environment."""
    torch = LazyImport("torch")
    rank = int(os.environ.get("RANK", rank))
    world_size = int(os.environ.get("WORLD_SIZE", world_size))
    if init_method is None:
        master_addr = os.environ.get("MASTER_ADDR", master_addr)
        master_port = os.environ.get("MASTER_PORT", master_port)
        init_method = 'env://{addr}:{port}'.format(addr=master_addr, port=master_port)
    torch.distributed.init_process_group(
        backend,
        init_method=init_method,
        world_size=world_size, 
        rank=rank
    )


def remove_label(input):
    if "labels" in input:  # for GLUE
        input.pop('labels')
    elif "start_positions" in input and "end_positions" in input:  # for SQuAD
        # pragma: no cover
        input.pop('start_positions')
        input.pop('end_positions')
    return input
